/*
 * Copyright 2020,2021,2022,2023 Sony Corporation
 */

#include <common.h>
#include <mapmem.h>
#include <time.h>
#include <dt_table.h>
#include <dm.h>
#include <debug_impl.h>
#include <mtk_dtbo.h>
#include <dtbo.h>
#include <linux/ctype.h>
#include <i2c.h>
#include <sony_fs.h>
#include <sony_panel.h>
#include <prodconfig.h>
#include <asm/gpio.h>
#include <adc.h>
#include <autodet.h>
#include <romtblo_impl.h>


static uint32_t pcd_idx;
static uint32_t pcd_i2c_idx;
static uint32_t pcd_adc_idx;
static uint32_t pcd_gpio_idx;
static uint32_t adc_idx;
static uint32_t uboot_adc_idx;
static uint32_t adc_i2c_idx;
static uint32_t gpio_idx;
static uint32_t gpio_i2c_idx;
static uint32_t i2c_idx;
static uint32_t panel_idx;
static uint32_t chassis_idx;
static uint32_t dt_map_idx;

static struct sony_pcd_rule_t *pcd_rules;
static struct sony_pcd_i2c_rule_t *pcd_i2c_rules;
static struct sony_pcd_adc_rule_t *pcd_adc_rules;
static struct sony_pcd_gpio_rule_t *pcd_gpio_rules;
static struct sony_adc_rule_t *adc_rules;
static struct sony_uboot_adc_rule_t *uboot_adc_rules;
static struct sony_adc_i2c_rule_t *adc_i2c_rules;
static struct sony_gpio_rule_t *gpio_rules;
static struct sony_gpio_i2c_rule_t *gpio_i2c_rules;
static struct sony_i2c_rule_t *i2c_rules;
static struct sony_panel_rule_t *panel_rules;
static struct sony_chassis_rule_t *chassis_rules;
static struct sony_dt_map_t *dt_map_info;

#define IC_PCB_CHIP_MAX			2
#define IC_PCB_CHIP_REVISION_MAX	4
#define IC_PCB_DTBO_MAX			4
#define I2C_LENGTH(x)			((x > 0xFFFFFF) ? 4 : (x > 0xFFFF) ? 3 : (x > 0xFF) ? 2 : 1)

static char *bootdata_list[] = {
	"sony/autodet/sony_pcd_det_rules.json",						// AUTODET_PCD
	"sony/autodet/sony_pcd_i2c_det_rules.json",					// AUTODET_PCD_I2C
	"sony/autodet/sony_pcd_adc_det_rules.json",					// AUTODET_PCD_ADC
	"sony/autodet/sony_pcd_gpio_det_rules.json",					// AUTODET_PCD_GPIO
	"sony/autodet/sony_adc_det_rules.json",						// AUTODET_ADC
	"sony/autodet/sony_uboot_adc_det_rules.json",					// AUTODET_UBOOT_ADC
	"sony/autodet/sony_adc_i2c_det_rules.json",					// AUTODET_ADC_I2C
	"sony/autodet/sony_gpio_det_rules.json",					// AUTODET_GPIO
	"sony/autodet/sony_gpio_i2c_det_rules.json",					// AUTODET_GPIO_I2C
	"sony/autodet/sony_i2c_det_rules.json",						// AUTODET_I2C
	"sony/autodet/sony_panel_det_rules.json",					// AUTODET_PANEL
	"sony/autodet/sony_chassis_det_rules.json",					// AUTODET_CHASSIS
	"sony/autodet/sony_dt_map.txt",							// AUTODET_DT_MAP
	NULL										// AUTODET_UNKNOWN
};

#define ALLOCATION(name, max_cat, max_item)													\
	({																	\
		if (name##_rules == NULL) {													\
			name##_rules = (struct sony_##name##_rule_t *)malloc(sizeof(struct sony_##name##_rule_t) * max_cat);			\
			assert(name##_rules != NULL);												\
			memset(name##_rules, 0x00, sizeof(struct sony_##name##_rule_t) * max_cat);						\
			for (i = 0; i < max_cat; i++) {												\
				name##_rules[i].items = (struct sony_##name##_item *)malloc(sizeof(struct sony_##name##_item) * max_item);	\
				assert(name##_rules[i].items != NULL);										\
				memset(name##_rules[i].items, 0x00, sizeof(struct sony_##name##_item) * max_item);				\
			}															\
		}																\
	})

#define FREE(name, max)							\
	({									\
		if (name##_rules != NULL) {					\
			for (i = 0; i < max; i++) {				\
				free(name##_rules[i].items);			\
				name##_rules[i].items = NULL;			\
			}							\
			free(name##_rules);					\
			name##_rules = NULL;					\
			name##_idx = 0;						\
		}								\
	})

static char *uboot_pcb_dtbo_list[][IC_PCB_DTBO_MAX][IC_PCB_CHIP_MAX] = {
	{
		{"uboot_mt5896",				"uboot_mt5897"},				/* uboot */
		{"uboot_MT5896_H2V1_B6_S",			"uboot_MT5897_H1V1_B6_S"},			/* uboot_pcb */
		{"rt9114_mt5896_h2v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			"ldm_128x128"},					/* ldm */
	},
	{
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* uboot */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* uboot_pcb */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* amp */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* ldm */
	},
	{
#if defined(CONFIG_ANDROID_CN_PLATFORM)
		{"uboot_mt5896_e3",				"uboot_mt5897_e3"},				/* uboot */
		{"uboot_MT5896_H1V4_B6_S",			"uboot_MT5897_H1V1_B6_S"},			/* uboot_pcb */
		{"rt9114_mt5896_h1v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#else
		{"uboot_mt5896_e3",				"uboot_mt5897_e3"},				/* uboot */
		{"uboot_MT5896_H2V3_B6_S",			"uboot_MT5897_H1V1_B6_S"},			/* uboot_pcb */
		{"rt9114_mt5896_h2v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#endif
	},
	{
#if defined(CONFIG_ANDROID_CN_PLATFORM)
		{"uboot_mt5896_e3",				"uboot_mt5897_e4"},				/* uboot */
		{"uboot_MT5896_H1V4_B6_S",			"uboot_MT5897_H1V1_B6_S"},			/* uboot_pcb */
		{"rt9114_mt5896_h1v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#else
		{"uboot_mt5896_e3",				"uboot_mt5897_e4"},				/* uboot */
		{"uboot_MT5896_H2V3_B6_S",			"uboot_MT5897_H1V1_B6_S"},			/* uboot_pcb */
		{"rt9114_mt5896_h2v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#endif
	}
};

static char *uboot_dtbo_list[][IC_PCB_CHIP_MAX] = {
	{"wakeup_base",					"wakeup_base"},					/* wakeup_key */
	{"vad_NULL_rproc",				"vad_NULL_rproc"},				/* vad */
	{"pmu_cm4_rproc",				"pmu_cm4_rproc"},				/* pmu */
#ifdef CONFIG_SONY_DTBO_DEFAULT_MMAP
	{"mmap_mt5896_h1v1_an_6g",			"mmap_mt5897_h1v1_an_4g"},			/* mmap */
#endif
	{SONY_DEFAULT_DTB_FILE,				SONY_DEFAULT_DTB_FILE},				/* Sony default */
	{SONY_DEFAULT_DTB_UBOOT_FILE,			SONY_DEFAULT_DTB_UBOOT_FILE},			/* Sony uboot specific */
	{"Sony_default_mt5896",				"Sony_default_mt5897"},				/* Sony PCBA specific */
	{"Sony_default_uboot_mt5896",			"Sony_default_uboot_mt5897"}			/* Sony uboot and PCBA specific */
};

static char *kernel_pcb_dtbo_list[][IC_PCB_DTBO_MAX][IC_PCB_CHIP_MAX] = {
	{
		{"mt5896",					"mt5897"},					/* ic_pcb */
		{"MT5896_H2V1_B6_S",				"MT5897_H1V1_B6_S"},				/* pcb */
		{"rt9114_mt5896_h2v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			"ldm_128x128"},					/* ldm */
	},
	{
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* uboot */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* uboot_pcb */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* amp */
		{SONY_DEFAULT_DTB_EMPTY_FILE,			SONY_DEFAULT_DTB_EMPTY_FILE},			/* ldm */
	},
	{
#if defined(CONFIG_ANDROID_CN_PLATFORM)
		{"mt5896_e3",					"mt5897_e3"},					/* ic_pcb */
		{"MT5896_H1V4_B6_S",				"MT5897_H1V1_B6_S"},				/* pcb */
		{"rt9114_mt5896_h1v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#else
		{"mt5896_e3",					"mt5897_e3"},					/* ic_pcb */
		{"MT5896_H2V3_B6_S",				"MT5897_H1V1_B6_S"},				/* pcb */
		{"rt9114_mt5896_h2v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#endif
	},
	{
#if defined(CONFIG_ANDROID_CN_PLATFORM)
		{"mt5896_e3",					"mt5897_e4"},					/* ic_pcb */
		{"MT5896_H1V4_B6_S",				"MT5897_H1V1_B6_S"},				/* pcb */
		{"rt9114_mt5896_h1v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#else
		{"mt5896_e3",					"mt5897_e4"},					/* ic_pcb */
		{"MT5896_H2V3_B6_S",				"MT5897_H1V1_B6_S"},				/* pcb */
		{"rt9114_mt5896_h2v1",				"rt9114_mt5897_h1v1"},				/* amp */
		{"ldm_128x128",					"ldm_128x128"},					/* ldm */
#endif
	}
};

static char *kernel_dtbo_list[][IC_PCB_CHIP_MAX] = {
	{"ir_base",					"ir_base"},					/* ir */
	{"keypad_base",					"keypad_base"},					/* keypad */
	{"vad_NULL_rproc",				"vad_NULL_rproc"},				/* vad */
	{"voc_mt58xx_normal",				"voc_mt58xx_normal"},				/* voc */
	{"pmu_cm4_rproc",				"pmu_cm4_rproc"},				/* pmu */
	{"audio_cpu_rproc",				"audio_cpu_rproc"},				/* audio_mcu */
	{"inputdev_cm4",				"inputdev_cm4"},				/* input */
#ifdef CONFIG_SONY_DTBO_DEFAULT_MMAP
	{"mmap_mt5896_h1v1_an_6g",			"mmap_mt5897_h1v1_an_4g"},			/* mmap */
	{"resv_mem_mt5896_h1v1_an_6g",			"resv_mem_mt5897_h1v1_an_4g"},			/* resv_mem */
	{"iommu_8k",					"iommu_4k"},					/* iommu */
	{"mt5896_6g",					"mt5897_4g"},					/* iommu_cust */
#endif
	{"bdma_base",					"bdma_base"},					/* bdma */
	{"mspi_base",					"mspi_base"},					/* mspi */
	{"wdt_base",					"wdt_base"},					/* wdt */
	{SONY_DEFAULT_DTB_FILE,				SONY_DEFAULT_DTB_FILE},				/* Sony default */
	{SONY_DEFAULT_DTB_KERNEL_FILE,			SONY_DEFAULT_DTB_KERNEL_FILE},			/* Sony kernel specific */
	{"Sony_default_mt5896",				"Sony_default_mt5897"},				/* Sony PCBA specific */
	{"Sony_default_kernel_mt5896",			"Sony_default_kernel_mt5897"}			/* Sony kernel and PCBA specific */
};

static int allocate_memory(void)
{
	int i = 0;

	ALLOCATION(pcd, MAX_PCD_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(pcd_i2c, MAX_PCD_I2C_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(pcd_adc, MAX_PCD_ADC_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(pcd_gpio, MAX_PCD_GPIO_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(adc, MAX_ADC_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(uboot_adc, MAX_UBOOT_ADC_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(adc_i2c, MAX_ADC_I2C_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(gpio, MAX_GPIO_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(gpio_i2c, MAX_GPIO_I2C_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(i2c, MAX_I2C_CATEGORY, MAX_AUTODET_ITEMS);
	ALLOCATION(panel, MAX_PANEL_CATEGORY, MAX_PANEL_ITEMS);
	ALLOCATION(chassis, MAX_CHASSIS_CATEGORY, MAX_CHASSIS_ITEMS);

	if (dt_map_info == NULL) {
		// allocate memory for dt_map_info
		dt_map_info = (struct sony_dt_map_t *)malloc(sizeof(struct sony_dt_map_t) * MAX_DT_MAP_ITEMS);
		assert(dt_map_info != NULL);
		memset(dt_map_info, 0x00, sizeof(struct sony_dt_map_t) * MAX_DT_MAP_ITEMS);
	}
	pcd_info_start();
	return 0;
}

static void free_memory(void)
{
	int i = 0;

	if (is_uboot_dtb())
		return;

	pcd_info_end();
	if (dt_map_info != NULL) {
		// free memory for dt_map_info
		free(dt_map_info);
		dt_map_info = NULL;
		dt_map_idx = 0;
	}
	FREE(chassis, MAX_CHASSIS_CATEGORY);
	FREE(panel, MAX_PANEL_CATEGORY);
	FREE(i2c, MAX_I2C_CATEGORY);
	FREE(gpio_i2c, MAX_GPIO_I2C_CATEGORY);
	FREE(gpio, MAX_GPIO_CATEGORY);
	FREE(adc_i2c, MAX_ADC_I2C_CATEGORY);
	FREE(uboot_adc, MAX_UBOOT_ADC_CATEGORY);
	FREE(adc, MAX_ADC_CATEGORY);
	FREE(pcd_gpio, MAX_PCD_GPIO_CATEGORY);
	FREE(pcd_adc, MAX_PCD_ADC_CATEGORY);
	FREE(pcd_i2c, MAX_PCD_I2C_CATEGORY);
	FREE(pcd, MAX_PCD_CATEGORY);
}

static int parse_rules_start(enum autodet_id id, const char *buf)
{
	char *find = NULL;
	int i = 0;
	char category[MAX_CATEGORY_LENGTH] = { 0 };

	// search '['(0x5b)
	find = strchr(buf, 0x5b);
	if (find == NULL)
		return 0;

	// search '"'(0x22)
	find = strchr(buf, 0x22);
	if (find != NULL) {
		while ((*(++find) != 0x22) && (*find != 0x00)) {
			category[i++] = *find;
			if (i == MAX_CATEGORY_LENGTH) {
				category[i-1] = '\0';
				break;
			}
		}
	}

	switch (id) {
	case AUTODET_PCD:
		memcpy(pcd_rules[pcd_idx].category, category, MAX_CATEGORY_LENGTH);
		if (strncmp(pcd_rules[pcd_idx].category, SONY_MULTI_KEY_PREFIX, strlen(SONY_MULTI_KEY_PREFIX)) == 0)
			pcd_rules[pcd_idx].is_multi_key = true;
		break;
	case AUTODET_PCD_I2C:
		memcpy(pcd_i2c_rules[pcd_i2c_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_PCD_ADC:
		memcpy(pcd_adc_rules[pcd_adc_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_PCD_GPIO:
		memcpy(pcd_gpio_rules[pcd_gpio_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_ADC:
		memcpy(adc_rules[adc_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_UBOOT_ADC:
		memcpy(uboot_adc_rules[uboot_adc_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_ADC_I2C:
		memcpy(adc_i2c_rules[adc_i2c_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_GPIO:
		memcpy(gpio_rules[gpio_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_GPIO_I2C:
		memcpy(gpio_i2c_rules[gpio_i2c_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_I2C:
		memcpy(i2c_rules[i2c_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_PANEL:
		memcpy(panel_rules[panel_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	case AUTODET_CHASSIS:
		memcpy(chassis_rules[chassis_idx].category, category, MAX_CATEGORY_LENGTH);
		break;
	default:
		UBOOT_ERROR("id = %d\n", id);
		return -1;
	}

	return 0;
}

static int parse_rules_end(enum autodet_id id, const char *buf)
{
	char *find = NULL;

	// search ']'(0x5d)
	find = strchr(buf, 0x5d);
	if (find == NULL)
		return 0;

	switch (id) {
	case AUTODET_PCD:
		if (pcd_rules[pcd_idx].category[0])
			pcd_idx++;
		break;
	case AUTODET_PCD_I2C:
		if (pcd_i2c_rules[pcd_i2c_idx].category[0])
			pcd_i2c_idx++;
		break;
	case AUTODET_PCD_ADC:
		if (pcd_adc_rules[pcd_adc_idx].category[0])
			pcd_adc_idx++;
		break;
	case AUTODET_PCD_GPIO:
		if (pcd_gpio_rules[pcd_gpio_idx].category[0])
			pcd_gpio_idx++;
		break;
	case AUTODET_ADC:
		if (adc_rules[adc_idx].category[0])
			adc_idx++;
		break;
	case AUTODET_UBOOT_ADC:
		if (uboot_adc_rules[uboot_adc_idx].category[0])
			uboot_adc_idx++;
		break;
	case AUTODET_ADC_I2C:
		if (adc_i2c_rules[adc_i2c_idx].category[0])
			adc_i2c_idx++;
		break;
	case AUTODET_GPIO:
		if (gpio_rules[gpio_idx].category[0])
			gpio_idx++;
		break;
	case AUTODET_GPIO_I2C:
		if (gpio_i2c_rules[gpio_i2c_idx].category[0])
			gpio_i2c_idx++;
		break;
	case AUTODET_I2C:
		if (i2c_rules[i2c_idx].category[0])
			i2c_idx++;
		break;
	case AUTODET_PANEL:
		if (panel_rules[panel_idx].category[0])
			panel_idx++;
		break;
	case AUTODET_CHASSIS:
		if (chassis_rules[chassis_idx].category[0])
			chassis_idx++;
		break;
	default:
		UBOOT_ERROR("id = %d\n", id);
		return -1;
	}

	return 0;
}

static uint32_t get_rule_num(enum autodet_id id, void *items)
{
	uint32_t num = 0, i = 0;

	switch (id) {
	case AUTODET_PCD:
		{
			struct sony_pcd_item *pcd = (struct sony_pcd_item *)items;

			for (i = 0; i < MAX_RULE_NUM; i++) {
				if ((pcd->key[i][0] != 0x00) && (pcd->value[i][0] != 0x00))
					num++;
				else
					break;
			}
		}
		break;
	case AUTODET_CHASSIS:
		{
			struct sony_chassis_item *chassis = (struct sony_chassis_item *)items;

			for (i = 0; i < MAX_RULE_NUM; i++) {
				if ((chassis->key[i][0] != 0x00) && (chassis->value[i][0] != 0x00))
					num++;
				else
					break;
			}
		}
		break;
	default:
		UBOOT_ERROR("id = %d\n", id);
		break;
	}
	return num;
}

static int set_pcd_item(const char *buf, int index)
{
	uint32_t cur_idx = pcd_rules[pcd_idx].count;
	uint32_t num = get_rule_num(AUTODET_PCD, (void *)&pcd_rules[pcd_idx].items[cur_idx]);

	if (num >= MAX_RULE_NUM) {
		UBOOT_ERROR("out of space in pcd item rule.\n");
		return 0;
	}
	if ((index % 2) == 0)
		memcpy(pcd_rules[pcd_idx].items[cur_idx].key[num], buf, MAX_KEY_LENGTH);

	if ((index % 2) == 1) {
		memcpy(pcd_rules[pcd_idx].items[cur_idx].value[num], buf, MAX_VALUE_LENGTH);
		if (strncmp(pcd_rules[pcd_idx].items[cur_idx].key[num], SONY_DEVICE_NAME, MAX_KEY_LENGTH) == 0)
			pcd_rules[pcd_idx].count++;
	}

	return 0;
}

static int set_pcd_i2c_item(const char *buf, int index)
{
	int cur_idx = pcd_i2c_rules[pcd_i2c_idx].count;

	switch (index) {
	case 0:
		memcpy(pcd_i2c_rules[pcd_i2c_idx].items[cur_idx].first_key, buf, MAX_KEY_LENGTH);
		break;
	case 1:
		memcpy(pcd_i2c_rules[pcd_i2c_idx].items[cur_idx].first_value, buf, MAX_VALUE_LENGTH);
		break;
	case 2:
		pcd_i2c_rules[pcd_i2c_idx].items[cur_idx].bus = strtoul(buf, NULL, 0);
		break;
	case 3:
		pcd_i2c_rules[pcd_i2c_idx].items[cur_idx].slave = strtoul(buf, NULL, 0);
		break;
	case 4:
		pcd_i2c_rules[pcd_i2c_idx].items[cur_idx].addr = strtoul(buf, NULL, 0);
		break;
	case 5:
		pcd_i2c_rules[pcd_i2c_idx].items[cur_idx].result = strtoul(buf, NULL, 0);
		break;
	case 6:
		memcpy(pcd_i2c_rules[pcd_i2c_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 6)
		pcd_i2c_rules[pcd_i2c_idx].count++;

	return 0;
}

static int set_pcd_adc_item(const char *buf, int index)
{
	int cur_idx = pcd_adc_rules[pcd_adc_idx].count;

	switch (index) {
	case 0:
		memcpy(pcd_adc_rules[pcd_adc_idx].items[cur_idx].first_key, buf, MAX_KEY_LENGTH);
		break;
	case 1:
		memcpy(pcd_adc_rules[pcd_adc_idx].items[cur_idx].first_value, buf, MAX_VALUE_LENGTH);
		break;
	case 2:
		pcd_adc_rules[pcd_adc_idx].items[cur_idx].channel = strtoul(buf, NULL, 0);
		break;
	case 3:
		pcd_adc_rules[pcd_adc_idx].items[cur_idx].min = strtoul(buf, NULL, 0);
		break;
	case 4:
		pcd_adc_rules[pcd_adc_idx].items[cur_idx].max = strtoul(buf, NULL, 0);
		break;
	case 5:
		memcpy(pcd_adc_rules[pcd_adc_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 5)
		pcd_adc_rules[pcd_adc_idx].count++;

	return 0;
}

static int set_pcd_gpio_item(const char *buf, int index)
{
	int cur_idx = pcd_gpio_rules[pcd_gpio_idx].count;

	switch (index) {
	case 0:
		memcpy(pcd_gpio_rules[pcd_gpio_idx].items[cur_idx].first_key, buf, MAX_KEY_LENGTH);
		break;
	case 1:
		memcpy(pcd_gpio_rules[pcd_gpio_idx].items[cur_idx].first_value, buf, MAX_VALUE_LENGTH);
		break;
	case 2:
		pcd_gpio_rules[pcd_gpio_idx].items[cur_idx].gpio = strtoul(buf, NULL, 0);
		break;
	case 3:
		pcd_gpio_rules[pcd_gpio_idx].items[cur_idx].result = strtoul(buf, NULL, 0);
		break;
	case 4:
		memcpy(pcd_gpio_rules[pcd_gpio_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 4)
		pcd_gpio_rules[pcd_gpio_idx].count++;

	return 0;
}

static int set_adc_item(const char *buf, int index)
{
	int cur_idx = adc_rules[adc_idx].count;

	switch (index) {
	case 0:
		adc_rules[adc_idx].items[cur_idx].channel = strtoul(buf, NULL, 0);
		break;
	case 1:
		adc_rules[adc_idx].items[cur_idx].min = strtoul(buf, NULL, 0);
		break;
	case 2:
		adc_rules[adc_idx].items[cur_idx].max = strtoul(buf, NULL, 0);
		break;
	case 3:
		memcpy(adc_rules[adc_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 3)
		adc_rules[adc_idx].count++;

	return 0;
}

static int set_uboot_adc_item(const char *buf, int index)
{
	int cur_idx = uboot_adc_rules[uboot_adc_idx].count;

	switch (index) {
	case 0:
		uboot_adc_rules[uboot_adc_idx].items[cur_idx].channel = strtoul(buf, NULL, 0);
		break;
	case 1:
		uboot_adc_rules[uboot_adc_idx].items[cur_idx].min = strtoul(buf, NULL, 0);
		break;
	case 2:
		uboot_adc_rules[uboot_adc_idx].items[cur_idx].max = strtoul(buf, NULL, 0);
		break;
	case 3:
		memcpy(uboot_adc_rules[uboot_adc_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 3)
		uboot_adc_rules[uboot_adc_idx].count++;

	return 0;
}

static int set_adc_i2c_item(const char *buf, int index)
{
	int cur_idx = adc_i2c_rules[adc_i2c_idx].count;

	switch (index) {
	case 0:
		adc_i2c_rules[adc_i2c_idx].items[cur_idx].channel = strtoul(buf, NULL, 0);
		break;
	case 1:
		adc_i2c_rules[adc_i2c_idx].items[cur_idx].min = strtoul(buf, NULL, 0);
		break;
	case 2:
		adc_i2c_rules[adc_i2c_idx].items[cur_idx].max = strtoul(buf, NULL, 0);
		break;
	case 3:
		adc_i2c_rules[adc_i2c_idx].items[cur_idx].bus = strtoul(buf, NULL, 0);
		break;
	case 4:
		adc_i2c_rules[adc_i2c_idx].items[cur_idx].slave = strtoul(buf, NULL, 0);
		break;
	case 5:
		adc_i2c_rules[adc_i2c_idx].items[cur_idx].addr = strtoul(buf, NULL, 0);
		break;
	case 6:
		adc_i2c_rules[adc_i2c_idx].items[cur_idx].result = strtoul(buf, NULL, 0);
		break;
	case 7:
		memcpy(adc_i2c_rules[adc_i2c_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 7)
		adc_i2c_rules[adc_i2c_idx].count++;

	return 0;
}

static int set_gpio_item(const char *buf, int index)
{
	int cur_idx = gpio_rules[gpio_idx].count;

	switch (index) {
	case 0:
		gpio_rules[gpio_idx].items[cur_idx].gpio = strtoul(buf, NULL, 0);
		break;
	case 1:
		gpio_rules[gpio_idx].items[cur_idx].result = strtoul(buf, NULL, 0);
		break;
	case 2:
		memcpy(gpio_rules[gpio_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 2)
		gpio_rules[gpio_idx].count++;

	return 0;
}

static int set_gpio_i2c_item(const char *buf, int index)
{
	int cur_idx = gpio_i2c_rules[gpio_i2c_idx].count;

	switch (index) {
	case 0:
		gpio_i2c_rules[gpio_i2c_idx].items[cur_idx].gpio = strtoul(buf, NULL, 0);
		break;
	case 1:
		gpio_i2c_rules[gpio_i2c_idx].items[cur_idx].result = strtoul(buf, NULL, 0);
		break;
	case 2:
		gpio_i2c_rules[gpio_i2c_idx].items[cur_idx].bus = strtoul(buf, NULL, 0);
		break;
	case 3:
		gpio_i2c_rules[gpio_i2c_idx].items[cur_idx].slave = strtoul(buf, NULL, 0);
		break;
	case 4:
		gpio_i2c_rules[gpio_i2c_idx].items[cur_idx].addr = strtoul(buf, NULL, 0);
		break;
	case 5:
		gpio_i2c_rules[gpio_i2c_idx].items[cur_idx].value = strtoul(buf, NULL, 0);
		break;
	case 6:
		memcpy(gpio_i2c_rules[gpio_i2c_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 6)
		gpio_i2c_rules[gpio_i2c_idx].count++;

	return 0;
}

static int set_i2c_item(const char *buf, int index)
{
	int cur_idx = i2c_rules[i2c_idx].count;

	switch (index) {
	case 0:
		i2c_rules[i2c_idx].items[cur_idx].bus = strtoul(buf, NULL, 0);
		break;
	case 1:
		i2c_rules[i2c_idx].items[cur_idx].slave = strtoul(buf, NULL, 0);
		break;
	case 2:
		i2c_rules[i2c_idx].items[cur_idx].addr = strtoul(buf, NULL, 0);
		break;
	case 3:
		i2c_rules[i2c_idx].items[cur_idx].result = strtoul(buf, NULL, 0);
		break;
	case 4:
		memcpy(i2c_rules[i2c_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 4)
		i2c_rules[i2c_idx].count++;

	return 0;
}

static uint8_t hex2digit(char hex)
{
	if (hex > 0x39)
		hex -= 7;

	return (hex & 0x0f);
}

static int set_panel_item(const char *buf, int index)
{
	int cur_idx = panel_rules[panel_idx].count;
	int i = 0;

	switch (index) {
	case 0:
		panel_rules[panel_idx].items[cur_idx].bus = strtoul(buf, NULL, 0);
		break;
	case 1:
		panel_rules[panel_idx].items[cur_idx].slave = strtoul(buf, NULL, 0);
		break;
	case 2:
		panel_rules[panel_idx].items[cur_idx].addr = strtoul(buf, NULL, 0);
		break;
	case 3:
		for (i = 0; i < MAX_PANELID_LENGTH; i++)
			panel_rules[panel_idx].items[cur_idx].panel_id[i] = hex2digit(buf[i * 2]) << 4 | hex2digit(buf[i * 2 + 1]);
		break;
	case 4:
		memcpy(panel_rules[panel_idx].items[cur_idx].device_name, buf, MAX_DEVICE_NAME_LENGTH);
		break;
	default:
		UBOOT_ERROR("index error(%d)\n", index);
		break;
	}
	if (index == 4)
		panel_rules[panel_idx].count++;

	return 0;
}

static int set_chassis_item(const char *buf, int index)
{
	uint32_t cur_idx = chassis_rules[chassis_idx].count;
	uint32_t num = get_rule_num(AUTODET_CHASSIS, (void *)&chassis_rules[chassis_idx].items[cur_idx]);

	if (num >= MAX_RULE_NUM) {
		UBOOT_ERROR("out of space in chassis item rule.\n");
		return 0;
	}
	if ((index % 2) == 0)
		memcpy(chassis_rules[chassis_idx].items[cur_idx].key[num], buf, MAX_KEY_LENGTH);

	if ((index % 2) == 1) {
		memcpy(chassis_rules[chassis_idx].items[cur_idx].value[num], buf, MAX_VALUE_LENGTH);
		if (strncmp(chassis_rules[chassis_idx].items[cur_idx].key[num], SONY_DEVICE_NAME, MAX_KEY_LENGTH) == 0)
			chassis_rules[chassis_idx].count++;
	}

	return 0;
}

static int set_rules_item(enum autodet_id id, const char *buf, int index)
{
	switch (id) {
	case AUTODET_PCD:
		set_pcd_item(buf, index);
		break;
	case AUTODET_PCD_I2C:
		set_pcd_i2c_item(buf, index);
		break;
	case AUTODET_PCD_ADC:
		set_pcd_adc_item(buf, index);
		break;
	case AUTODET_PCD_GPIO:
		set_pcd_gpio_item(buf, index);
		break;
	case AUTODET_ADC:
		set_adc_item(buf, index);
		break;
	case AUTODET_UBOOT_ADC:
		set_uboot_adc_item(buf, index);
		break;
	case AUTODET_ADC_I2C:
		set_adc_i2c_item(buf, index);
		break;
	case AUTODET_GPIO:
		set_gpio_item(buf, index);
		break;
	case AUTODET_GPIO_I2C:
		set_gpio_i2c_item(buf, index);
		break;
	case AUTODET_I2C:
		set_i2c_item(buf, index);
		break;
	case AUTODET_PANEL:
		set_panel_item(buf, index);
		break;
	case AUTODET_CHASSIS:
		set_chassis_item(buf, index);
		break;
	default:
		UBOOT_ERROR("id = %d\n", id);
		return -1;
	}
	return 0;
}

static int parse_rules_item(enum autodet_id id, const char *buf, bool flag)
{
	char *find = NULL;
	char value[MAX_VALUE_LENGTH] = { 0 };
	int i = 0;
	int index = 0;

	// search '{'(0x7b) and '}'(0x7d)
	find = strchr(buf, 0x7b);
	if (find == NULL)
		return -1;
	find = strchr(buf, 0x7d);
	if (find == NULL)
		return -1;

	while (*buf != 0x00) {
		if (!flag) {
			// search ':'(0x3a)
			find = strchr(buf, 0x3a);
			if (find != NULL)
				buf = find;
		}
		// search '"'(0x22)
		find = strchr(buf, 0x22);
		if (find != NULL) {
			i = 0;
			memset(value, 0x00, MAX_VALUE_LENGTH);
			// search '"'(0x22)
			while ((*(++find) != 0x22) && (*find != 0x00)) {
				value[i++] = *find;
				if (i == MAX_VALUE_LENGTH) {
					value[i-1] = '\0';
					break;
				}
			}
			if (value[0] != 0x00)
				set_rules_item(id, value, index++);

			switch (id) {
			case AUTODET_PCD_I2C:
			case AUTODET_PCD_ADC:
			case AUTODET_PCD_GPIO:
				if (index > 1)
					flag = false;
				break;
			default:
				break;
			}
			buf = ++find;
		}
		buf++;
	}
	return 0;
}

static int parse_rules_data(enum autodet_id id, const char *buf)
{
	char readline[512] = { 0 };
	char *find = NULL;
	bool offset = false, enabled = true;

	while (*buf != 0x00) {
		// search 'LF'(0x0a)
		find = strchr(buf, 0x0a);
		if (find != NULL) {
			memset(readline, 0x00, sizeof(readline));
			memcpy(readline, buf, (find - buf));
			buf += strlen(readline);
		}

		switch (id) {
		case AUTODET_PCD:
			if (!((pcd_idx < MAX_PCD_CATEGORY) && (pcd_rules[pcd_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in pcd_rules.\n");
			}
			offset = true;
			break;
		case AUTODET_PCD_I2C:
			if (!((pcd_i2c_idx < MAX_PCD_I2C_CATEGORY) && (pcd_i2c_rules[pcd_i2c_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in pcd_i2c_rules.\n");
			}
			offset = true;
			break;
		case AUTODET_PCD_ADC:
			if (!((pcd_adc_idx < MAX_PCD_ADC_CATEGORY) && (pcd_adc_rules[pcd_adc_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in pcd_adc_rules.\n");
			}
			offset = true;
			break;
		case AUTODET_PCD_GPIO:
			if (!((pcd_gpio_idx < MAX_PCD_GPIO_CATEGORY) && (pcd_gpio_rules[pcd_gpio_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in pcd_gpio_rules.\n");
			}
			offset = true;
			break;
		case AUTODET_ADC:
			if (!((adc_idx < MAX_ADC_CATEGORY) && (adc_rules[adc_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in adc_rules.\n");
			}
			break;
		case AUTODET_UBOOT_ADC:
			if (!((uboot_adc_idx < MAX_UBOOT_ADC_CATEGORY) && (uboot_adc_rules[uboot_adc_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in uboot_adc_rules.\n");
			}
			break;
		case AUTODET_ADC_I2C:
			if (!((adc_i2c_idx < MAX_ADC_I2C_CATEGORY) && (adc_i2c_rules[adc_i2c_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in adc_i2c_rules.\n");
			}
			break;
		case AUTODET_GPIO:
			if (!((gpio_idx < MAX_GPIO_CATEGORY) && (gpio_rules[gpio_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in gpio_rules.\n");
			}
			break;
		case AUTODET_GPIO_I2C:
			if (!((gpio_i2c_idx < MAX_GPIO_I2C_CATEGORY) && (gpio_i2c_rules[gpio_i2c_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in gpio_i2c_rules.\n");
			}
			break;
		case AUTODET_I2C:
			if (!((i2c_idx < MAX_I2C_CATEGORY) && (i2c_rules[i2c_idx].count < MAX_AUTODET_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in i2c_rules.\n");
			}
			break;
		case AUTODET_PANEL:
			if (!((panel_idx < MAX_PANEL_CATEGORY) && (panel_rules[panel_idx].count < MAX_PANEL_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in panel_rules.\n");
			}
			break;
		case AUTODET_CHASSIS:
			if (!((chassis_idx < MAX_CHASSIS_CATEGORY) && (chassis_rules[chassis_idx].count < MAX_CHASSIS_ITEMS))) {
				enabled = false;
				UBOOT_ERROR("out of space in chassis_rules.\n");
			}
			offset = true;
			break;
		default:
			UBOOT_ERROR("id = %d\n", id);
			return -1;
		}
		if (enabled) {
			parse_rules_start(id, readline);
			parse_rules_item(id, readline, offset);
			parse_rules_end(id, readline);
		}
		buf++;
	}
	return 0;
}

static int parse_dt_map_data(const char *buf)
{
	char *find = NULL;
	char readline[128] = { 0 };
	int len = 0;

	while (*buf != 0x00) {
		if (dt_map_idx == MAX_DT_MAP_ITEMS) {
			printf("Error: out of space in dt_map_info.\n");
			return -1;
		}
		// skip line starting with "["
		if (*buf == 0x5b) {
			buf = strchr(buf, 0x0a);
			++buf;
		}
		// search 'LF'(0x0a)
		find = strchr(buf, 0x0a);
		if (find != NULL) {
			len = find - buf;
			if (len > sizeof(readline)) {
				len = sizeof(readline);
				UBOOT_DEBUG("buffer size over.\n");
			}
			memset(readline, 0x00, sizeof(readline));
			memcpy(readline, buf, len);
			buf += len;
		}
		// search '='(0x3d)
		find = strchr(readline, 0x3d);
		if (find != NULL) {
			memcpy(dt_map_info[dt_map_idx].device_name, readline, find - readline);
			++find;
			dt_map_info[dt_map_idx].dtb_entry = strtoul(find, NULL, 10);
		}
		if (dt_map_info[dt_map_idx].device_name[0] != 0x00)
			dt_map_idx++;
		buf++;
	}
	return 0;
}

static int sony_dtb_overlay(const char *device_name)
{
	int i;
	int ret;

	for (i = 0; i < dt_map_idx; i++) {
		if (strncmp(dt_map_info[i].device_name, device_name, MAX_DEVICE_NAME_LENGTH) == 0) {
			ret = dtb_overlay(dt_map_info[i].dtb_entry);
			if (ret != 0) {
				UBOOT_ERROR("dtb_overlay %d: %s failed.\n", dt_map_info[i].dtb_entry, dt_map_info[i].device_name);
				return -1;
			}
			if (!is_uboot_dtb())
				add_dtbo_idx(dt_map_info[i].dtb_entry);
			printf("dtb_overlay %d: %s done.\n", dt_map_info[i].dtb_entry, dt_map_info[i].device_name);

			return 0;
		}
	}
	printf("dtb_overlay: %s not found.\n", device_name);

	return -1;
}

static int check_i2c_status(int bus, int slave, int addr, int len)
{
	uint8_t buf[5] = { 0 };
	struct udevice *dev = NULL;

	if (i2c_get_chip_for_busnum(bus, slave, 1, &dev) != 0)
		return -1;
	if ((addr == 0) && (len == 1)) {
		// To support the valid devices which don't have a unique identification data.
		return 0;
	}
	if (dm_i2c_read(dev, addr, buf, len) != 0)
		return -1;
	if (len == 1)
		return (int)buf[0];
	else if (len == 2)
		return (int)((buf[0] << 8) | buf[1]);
	else if (len == 3)
		return (int)((buf[0] << 16) | (buf[1] << 8) | buf[2]);
	else if (len == 4)
		return (int)((buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]);

	return -1;
}

static int check_panel_status(int bus, int slave, int addr, uint8_t *panel_id)
{
	uint8_t buf[MAX_PANELID_LENGTH + 1] = { 0 };
	struct udevice *dev = NULL;
	uint32_t retry = MAX_PANEL_RETRY_COUNT;
	uint32_t valid_count = 0;

	if (i2c_get_chip_for_busnum(bus, slave, 2, &dev) != 0)
		return -1;

	while (retry-- > 0) {
		if (dm_i2c_read(dev, addr, buf, MAX_PANELID_LENGTH) != 0) {
			valid_count = 0;
			continue;
		} else
			valid_count++;

		if (valid_count == 1) {
			// Copy 1st read data.
			memcpy(panel_id, buf, MAX_PANELID_LENGTH);
			memset(buf, 0, sizeof(buf));
		} else if (valid_count == 2) {
			valid_count = 0;
			// Compare 1st read data and 2nd ones.
			if (memcmp(panel_id, buf, MAX_PANELID_LENGTH) == 0)
				return 0;
		}
	}

	return -1;
}

static int check_gpio_status(int gpio)
{
	int ret = -1;
	int value = -1;

	ret = gpio_request(gpio, "audodet");
	if ((ret) && (ret != -EBUSY)) {
		UBOOT_ERROR("gpio: requesting pin %u failed ret=%d\n", gpio, ret);
		return -1;
	}
	gpio_direction_input(gpio);
	value = gpio_get_value(gpio);

	if (ret != -EBUSY)
		gpio_free(gpio);

	return value;
}

static int check_adc_status(int adc_ch)
{
	int ret = -1;
	struct udevice *dev = NULL;
	uint32_t data = 0;

	ret = uclass_get_device(UCLASS_ADC, 0, &dev);
	if (ret) {
		UBOOT_ERROR("Cannot find adc device: %d\n", ret);
		return ret;
	}
	ret = adc_start_channel(dev, adc_ch);
	if (ret) {
		UBOOT_ERROR("adc start channel error: %d\n", ret);
		return ret;
	}
	ret = adc_channel_data(dev, adc_ch, &data);
	if (ret) {
		UBOOT_ERROR("adc channel data error: %d\n", ret);
		return ret;
	}
	return data;
}

static int prepare_device_probe(void)
{
	int ret = -1;
	struct udevice *dev = NULL;
	// This is workaround code that ADC GPIO overlay of u-boot dtbo process failed
	// after select_sony_device_default_overlay().
	if (is_uboot_dtb()) {
		ret = uclass_get_device(UCLASS_ADC, 0, &dev);
		ret = uclass_get_device(UCLASS_GPIO, 0, &dev);
	}
	return ret;
}

static int open_and_parser(enum autodet_id id)
{
	char *buf = NULL;
	char *filename = NULL;
	int size = -1;

	if (id >= AUTODET_UNKNOWN) {
		UBOOT_ERROR("autodet id error.(%d)\n", id);
		return -1;
	}

	filename = bootdata_list[id];
	size = ext4_f_size(BOOTDATA_PART_NAME, filename);
	if (size == -1) {
		UBOOT_DEBUG("file size error.\n");
		return -1;
	}

	buf = malloc(size+1);
	if (buf == NULL) {
		UBOOT_ERROR("malloc error\n");
		return -1;
	}

	memset(buf, 0, size+1);
	if (ext4_f_load(BOOTDATA_PART_NAME, filename, buf) > 0) {
		switch (id) {
		case AUTODET_PCD:
		case AUTODET_PCD_I2C:
		case AUTODET_PCD_ADC:
		case AUTODET_PCD_GPIO:
		case AUTODET_ADC:
		case AUTODET_UBOOT_ADC:
		case AUTODET_ADC_I2C:
		case AUTODET_GPIO:
		case AUTODET_GPIO_I2C:
		case AUTODET_I2C:
		case AUTODET_PANEL:
		case AUTODET_CHASSIS:
			parse_rules_data(id, (const char *)buf);
			break;
		case AUTODET_DT_MAP:
			parse_dt_map_data((const char *)buf);
			break;
		default:
			UBOOT_ERROR("id = %d\n", id);
			break;
		}
	}
	free(buf);
	return 0;
}

static int get_bootdata(void)
{
	int ret = -1;

	if (!is_uboot_dtb())
		return 0;

	// open file and parser json files
	if (pcd_idx == 0) {
		ret = open_and_parser(AUTODET_PCD);
		if (ret != 0)
			UBOOT_INFO("pcd file open error. ret=%d\n", ret);
	}

	if (pcd_i2c_idx == 0) {
		ret = open_and_parser(AUTODET_PCD_I2C);
		if (ret != 0)
			UBOOT_INFO("pcd i2c file open error. ret=%d\n", ret);
	}

	if (pcd_adc_idx == 0) {
		ret = open_and_parser(AUTODET_PCD_ADC);
		if (ret != 0)
			UBOOT_INFO("pcd adc file open error. ret=%d\n", ret);
	}

	if (pcd_gpio_idx == 0) {
		ret = open_and_parser(AUTODET_PCD_GPIO);
		if (ret != 0)
			UBOOT_INFO("pcd gpio file open error. ret=%d\n", ret);
	}

	if (adc_idx == 0) {
		ret = open_and_parser(AUTODET_ADC);
		if (ret != 0)
			UBOOT_INFO("adc file open error. ret=%d\n", ret);
	}

	if (uboot_adc_idx == 0) {
		ret = open_and_parser(AUTODET_UBOOT_ADC);
		if (ret != 0)
			UBOOT_INFO("uboot adc file open error. ret=%d\n", ret);
	}

	if (adc_i2c_idx == 0) {
		ret = open_and_parser(AUTODET_ADC_I2C);
		if (ret != 0)
			UBOOT_INFO("adc i2c file open error. ret=%d\n", ret);
	}

	if (gpio_idx == 0) {
		ret = open_and_parser(AUTODET_GPIO);
		if (ret != 0)
			UBOOT_INFO("gpio file open error. ret=%d\n", ret);
	}

	if (gpio_i2c_idx == 0) {
		ret = open_and_parser(AUTODET_GPIO_I2C);
		if (ret != 0)
			UBOOT_INFO("gpio i2c file open error. ret=%d\n", ret);
	}

	if (i2c_idx == 0) {
		ret = open_and_parser(AUTODET_I2C);
		if (ret != 0)
			UBOOT_INFO("i2c file open error. ret=%d\n", ret);
	}

	if (panel_idx == 0) {
		ret = open_and_parser(AUTODET_PANEL);
		if (ret != 0)
			UBOOT_INFO("panel file open error. ret=%d\n", ret);
	}

	if (chassis_idx == 0) {
		ret = open_and_parser(AUTODET_CHASSIS);
		if (ret != 0)
			UBOOT_INFO("chassis file open error. ret=%d\n", ret);
	}

	if (dt_map_idx == 0) {
		ret = open_and_parser(AUTODET_DT_MAP);
		if (ret != 0)
			UBOOT_INFO("device map file open error. ret=%d\n", ret);
	}

	return 0;
}

static int select_sony_overlay_for_kernel(void)
{
	uint64_t start = 0, end = 0;
	uint32_t i = 0, j = 0, k = 0, num = 0;
	int adc_value = -1, i2c_value = -1, gpio_value = -1;
	bool i2c_access = true, do_overlay = true, check_pcd = true;
	char first_value[MAX_VALUE_LENGTH] = { 0 };
	char panel_id[MAX_PANELID_LENGTH + 1] = { 0 };
	char pcd_value[MAX_RULE_NUM][MAX_VALUE_LENGTH] = { 0 };

	start = get_timer(0);

	// select CHASSIS
	for (i = 0; i < chassis_idx; i++) {
		for (j = 0; j < chassis_rules[i].count; j++) {
			do_overlay = true;
			// get pcd data with first item rule.
			if (j == 0) {
				num = get_rule_num(AUTODET_CHASSIS, (void *)&chassis_rules[i].items[j]);
				if (num == 0) {
					UBOOT_DEBUG("failed to get chassis rule on %s\n", chassis_rules[i].category);
					break;
				}
				memset(pcd_value, 0, sizeof(char) * MAX_RULE_NUM * MAX_VALUE_LENGTH);
				check_pcd = true;
				for (k = 0; k < (num-1); k++) {
					if (get_product_config_value(chassis_rules[i].items[j].key[k], pcd_value[k]) != 0) {
						check_pcd = false;
						break;
					}
				}
				if (!check_pcd) {
					UBOOT_DEBUG("failed to get pcd data on %s\n", chassis_rules[i].category);
					break;
				}
			}
			// compare rule data and pcd data.
			for (k = 0; k < (num-1); k++) {
				if (strncmp(chassis_rules[i].items[j].value[k], pcd_value[k], MAX_VALUE_LENGTH) != 0) {
					do_overlay = false;
					break;
				}
			}
			if (do_overlay) {
				sony_dtb_overlay(chassis_rules[i].items[j].value[num-1]);
				break;
			}
		}
	}

	// select PCD
	for (i = 0; i < pcd_idx; i++) {
		// in case of PMU, autodet should not execute on kernel.
		if (strstr(pcd_rules[i].category, SONY_FOR_PMU_SUFFIX) != NULL)
			continue;
		for (j = 0; j < pcd_rules[i].count; j++) {
			do_overlay = true;
			// get pcd data with first item rule.
			if (j == 0) {
				num = get_rule_num(AUTODET_PCD, (void *)&pcd_rules[i].items[j]);
				if (num == 0) {
					UBOOT_DEBUG("failed to get pcd rule on %s\n", pcd_rules[i].category);
					break;
				}
				memset(pcd_value, 0, sizeof(char) * MAX_RULE_NUM * MAX_VALUE_LENGTH);
				check_pcd = true;
				for (k = 0; k < (num-1); k++) {
					if (get_product_config_value(pcd_rules[i].items[j].key[k], pcd_value[k]) != 0) {
						check_pcd = false;
						break;
					}
				}
				if (!check_pcd) {
					UBOOT_DEBUG("failed to get pcd data on %s\n", pcd_rules[i].category);
					break;
				}
			}
			// compare rule data and pcd data.
			for (k = 0; k < (num-1); k++) {
				if (strncmp(pcd_rules[i].items[j].value[k], pcd_value[k], MAX_VALUE_LENGTH) != 0) {
					do_overlay = false;
					break;
				}
			}
			if (do_overlay) {
				sony_dtb_overlay(pcd_rules[i].items[j].value[num-1]);
				break;
			}
		}
	}

	// select PCD I2C
	for (i = 0; i < pcd_i2c_idx; i++) {
		memset(first_value, 0x00, sizeof(first_value));
		if (get_product_config_value(pcd_i2c_rules[i].items[0].first_key, first_value) != 0)
			continue;

		for (j = 0; j < pcd_i2c_rules[i].count; j++) {
			i2c_access = true;
			if (j > 0) {
				if ((pcd_i2c_rules[i].items[j].bus == pcd_i2c_rules[i].items[j-1].bus) &&
					(pcd_i2c_rules[i].items[j].slave == pcd_i2c_rules[i].items[j-1].slave) &&
					(pcd_i2c_rules[i].items[j].addr == pcd_i2c_rules[i].items[j-1].addr))
					i2c_access = false;
			}
			if (i2c_access) {
				i2c_value = check_i2c_status(pcd_i2c_rules[i].items[j].bus,
											pcd_i2c_rules[i].items[j].slave,
											pcd_i2c_rules[i].items[j].addr,
											I2C_LENGTH(pcd_i2c_rules[i].items[j].result));
				if (i2c_value == -1)
					continue;

			}
			if ((strncmp(pcd_i2c_rules[i].items[j].first_value, first_value, MAX_VALUE_LENGTH) == 0) &&
				(pcd_i2c_rules[i].items[j].result == i2c_value)) {
				sony_dtb_overlay(pcd_i2c_rules[i].items[j].device_name);
				break;
			}
		}
	}

	// select PCD ADC
	for (i = 0; i < pcd_adc_idx; i++) {
		memset(first_value, 0x00, sizeof(first_value));
		if (get_product_config_value(pcd_adc_rules[i].items[0].first_key, first_value) != 0)
			continue;

		adc_value = check_adc_status(pcd_adc_rules[i].items[0].channel);
		if (adc_value == -1)
			continue;

		for (j = 0; j < pcd_adc_rules[i].count; j++) {
			if ((strncmp(pcd_adc_rules[i].items[j].first_value, first_value, MAX_VALUE_LENGTH) == 0) &&
				(pcd_adc_rules[i].items[j].min <= adc_value) &&
				(pcd_adc_rules[i].items[j].max >= adc_value)) {
				sony_dtb_overlay(pcd_adc_rules[i].items[j].device_name);
				break;
			}
		}
	}

	// select PCD GPIO
	for (i = 0; i < pcd_gpio_idx; i++) {
		memset(first_value, 0x00, sizeof(first_value));
		if (get_product_config_value(pcd_gpio_rules[i].items[0].first_key, first_value) != 0)
			continue;

		for (j = 0; j < pcd_gpio_rules[i].count; j++) {
			gpio_value = check_gpio_status(pcd_gpio_rules[i].items[j].gpio);
			if (gpio_value == -1)
				continue;

			if ((strncmp(pcd_gpio_rules[i].items[j].first_value, first_value, MAX_VALUE_LENGTH) == 0) &&
				(pcd_gpio_rules[i].items[j].result == gpio_value)) {
				sony_dtb_overlay(pcd_gpio_rules[i].items[j].device_name);
				break;
			}
		}
	}

	// select ADC
	for (i = 0; i < adc_idx; i++) {
		adc_value = check_adc_status(adc_rules[i].items[0].channel);
		if (adc_value == -1)
			continue;

		for (j = 0; j < adc_rules[i].count; j++) {
			if ((adc_rules[i].items[j].min <= adc_value) &&
				(adc_rules[i].items[j].max >= adc_value)) {
				sony_dtb_overlay(adc_rules[i].items[j].device_name);
				break;
			}
		}
	}

	// select ADC I2C
	for (i = 0; i < adc_i2c_idx; i++) {
		adc_value = check_adc_status(adc_i2c_rules[i].items[0].channel);
		if (adc_value == -1)
			continue;

		for (j = 0; j < adc_i2c_rules[i].count; j++) {
			i2c_access = true;
			if (j > 0) {
				if ((adc_i2c_rules[i].items[j].bus == adc_i2c_rules[i].items[j-1].bus) &&
					(adc_i2c_rules[i].items[j].slave == adc_i2c_rules[i].items[j-1].slave) &&
					(adc_i2c_rules[i].items[j].addr == adc_i2c_rules[i].items[j-1].addr))
					i2c_access = false;
			}
			if (i2c_access) {
				i2c_value = check_i2c_status(adc_i2c_rules[i].items[j].bus,
											adc_i2c_rules[i].items[j].slave,
											adc_i2c_rules[i].items[j].addr,
											I2C_LENGTH(adc_i2c_rules[i].items[j].result));
				if (i2c_value == -1)
					continue;

			}
			if ((adc_i2c_rules[i].items[j].min <= adc_value) &&
				(adc_i2c_rules[i].items[j].max >= adc_value) &&
				(adc_i2c_rules[i].items[j].result == i2c_value)) {
				sony_dtb_overlay(adc_i2c_rules[i].items[j].device_name);
				break;
			}
		}
	}

	// select GPIO
	for (i = 0; i < gpio_idx; i++) {
		for (j = 0; j < gpio_rules[i].count; j++) {
			gpio_value = check_gpio_status(gpio_rules[i].items[j].gpio);
			if (gpio_value == -1)
				continue;

			if (gpio_rules[i].items[j].result == gpio_value)
				sony_dtb_overlay(gpio_rules[i].items[j].device_name);
		}
	}

	// select GPIO I2C
	for (i = 0; i < gpio_i2c_idx; i++) {
		for (j = 0; j < gpio_i2c_rules[i].count; j++) {
			gpio_value = check_gpio_status(gpio_i2c_rules[i].items[j].gpio);
			if (gpio_value == -1)
				continue;

			i2c_access = true;
			if (j > 0) {
				if ((gpio_i2c_rules[i].items[j].bus == gpio_i2c_rules[i].items[j-1].bus) &&
					(gpio_i2c_rules[i].items[j].slave == gpio_i2c_rules[i].items[j-1].slave) &&
					(gpio_i2c_rules[i].items[j].addr == gpio_i2c_rules[i].items[j-1].addr))
					i2c_access = false;
			}
			if (i2c_access) {
				i2c_value = check_i2c_status(gpio_i2c_rules[i].items[j].bus,
											gpio_i2c_rules[i].items[j].slave,
											gpio_i2c_rules[i].items[j].addr,
											I2C_LENGTH(gpio_i2c_rules[i].items[j].result));
				if (i2c_value == -1)
					continue;

			}
			if ((gpio_i2c_rules[i].items[j].result == gpio_value) &&
				(gpio_i2c_rules[i].items[j].value == i2c_value)) {
				sony_dtb_overlay(gpio_i2c_rules[i].items[j].device_name);
				break;
			}
		}
	}

	// select I2C
	for (i = 0; i < i2c_idx; i++) {
		for (j = 0; j < i2c_rules[i].count; j++) {
			i2c_access = true;
			if (j > 0) {
				if ((i2c_rules[i].items[j].bus == i2c_rules[i].items[j-1].bus) &&
					(i2c_rules[i].items[j].slave == i2c_rules[i].items[j-1].slave) &&
					(i2c_rules[i].items[j].addr == i2c_rules[i].items[j-1].addr))
					i2c_access = false;
			}
			if (i2c_access) {
				i2c_value = check_i2c_status(i2c_rules[i].items[j].bus,
											i2c_rules[i].items[j].slave,
											i2c_rules[i].items[j].addr,
											I2C_LENGTH(i2c_rules[i].items[j].result));
				if (i2c_value == -1)
					continue;

			}
			if (i2c_rules[i].items[j].result == i2c_value) {
				sony_dtb_overlay(i2c_rules[i].items[j].device_name);
				break;
			}
		}
	}

	// select Panel
	for (i = 0; i < panel_idx; i++) {
		for (j = 0; j < panel_rules[i].count; j++) {
			paneldet_getPanelId(panel_id);
			if (memcmp(panel_rules[i].items[j].panel_id, panel_id, MAX_PANELID_LENGTH) == 0) {
				sony_dtb_overlay(panel_rules[i].items[j].device_name);
				break;
			}
		}
	}

	end = get_timer(0);
	UBOOT_BOOTTIME("\n[%s:%d][start:%llu][end:%llu][total time:%llu]\n", __func__, __LINE__, start, end, end - start);
	return 0;
}

static int select_sony_device_default_overlay(void)
{
	int i;
	uint32_t id, revision, rev_index;
	char *(*dtbo_list)[IC_PCB_CHIP_MAX];
	uint32_t num_dtbo;

	id = romtbl_get_chip_id_info();
	revision = romtbl_get_chip_revision_info();
	rev_index = revision - 1;
	if ((id >= IC_PCB_CHIP_MAX) || (rev_index >= IC_PCB_CHIP_REVISION_MAX)) {
		printf("Wrong ChipID (%d:%d). Set default ChipID %d:%d\n", id, revision, 0, 1);
		id = 0;
		rev_index = 0;
	}
	if (is_uboot_dtb()) {
		dtbo_list = uboot_pcb_dtbo_list[rev_index];
		num_dtbo = sizeof(uboot_pcb_dtbo_list[rev_index]) / sizeof(*uboot_pcb_dtbo_list[rev_index]);
	} else {
		dtbo_list = kernel_pcb_dtbo_list[rev_index];
		num_dtbo = sizeof(kernel_pcb_dtbo_list[rev_index]) / sizeof(*kernel_pcb_dtbo_list[rev_index]);
	}

	/* overlay pcb */
	for (i = 0; i < num_dtbo; i++) {
		if (strncmp(dtbo_list[i][id], SONY_DEFAULT_DTB_EMPTY_FILE, strlen(SONY_DEFAULT_DTB_EMPTY_FILE)) == 0)
			continue;
		sony_dtb_overlay(dtbo_list[i][id]);
	}

	if (is_uboot_dtb()) {
		dtbo_list = uboot_dtbo_list;
		num_dtbo = sizeof(uboot_dtbo_list) / sizeof(*uboot_dtbo_list);
	} else {
		dtbo_list = kernel_dtbo_list;
		num_dtbo = sizeof(kernel_dtbo_list) / sizeof(*kernel_dtbo_list);
	}

	/* add default dtb entries which always overlay */
	for (i = 0; i < num_dtbo; i++) {
		if (strncmp(dtbo_list[i][id], SONY_DEFAULT_DTB_EMPTY_FILE, strlen(SONY_DEFAULT_DTB_EMPTY_FILE)) == 0)
			continue;
		sony_dtb_overlay(dtbo_list[i][id]);
	}

	return 0;
}

int sony_dtbo_selection_and_overlay(void)
{
	uint64_t start = 0, end = 0;

	start = get_timer(0);

	// allocate memory for autodet files
	allocate_memory();

	// get json and map files from bootdata
	get_bootdata();

	// prepare device probe
	prepare_device_probe();

	/* cerate dtb entry list to overlay */
	select_sony_device_default_overlay();
	if (is_uboot_dtb()) {
		select_sony_overlay_for_uboot();
		select_sony_pcd_overlay_for_pmu();
	} else {
		select_sony_overlay_for_kernel();
	}

	// doesn't free memory on u-boot
	free_memory();

	end = get_timer(0);
	UBOOT_BOOTTIME("\n[%s:%d][start:%llu][end:%llu][total time:%llu]\n", __func__, __LINE__, start, end, end - start);
	return 0;
}

int select_sony_pcd_overlay(const char *category)
{
	int i = 0, j = 0, k = 0, num = 0;
	bool do_overlay = true;
	char pcd_value[MAX_RULE_NUM][MAX_VALUE_LENGTH] = { 0 };

	// select devices without HW access
	for (i = 0; i < pcd_idx; i++) {
		if (strncmp(pcd_rules[i].category, category, MAX_CATEGORY_LENGTH) != 0)
			continue;
		for (j = 0; j < pcd_rules[i].count; j++) {
			do_overlay = true;
			if (j == 0) {
				num = get_rule_num(AUTODET_PCD, (void *)&pcd_rules[i].items[j]);
				if (num == 0) {
					UBOOT_ERROR("failed to get number of keys on %s\n", pcd_rules[i].category);
					return AUTODET_NO_DATA;
				}
				for (k = 0; k < (num-1); k++) {
					// get product config value with key array.
					if (get_product_config_value(pcd_rules[i].items[j].key[k], pcd_value[k]) != 0) {
						UBOOT_ERROR("failed to get key data.(%s)\n", pcd_rules[i].items[j].key[k]);
						return AUTODET_NO_DATA;
					}
				}
			}
			// compare rule data and pcd data.
			for (k = 0; k < (num-1); k++) {
				if (strncmp(pcd_rules[i].items[j].value[k], pcd_value[k], MAX_VALUE_LENGTH) != 0) {
					do_overlay = false;
					break;
				}
			}
			if (do_overlay)
				return sony_dtb_overlay(pcd_rules[i].items[j].value[num-1]);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_chassis_overlay(const char *category)
{
	int i = 0, j = 0, k = 0, num = 0;
	bool do_overlay = true;
	char pcd_value[MAX_RULE_NUM][MAX_VALUE_LENGTH] = { 0 };

	// select devices without HW access
	for (i = 0; i < chassis_idx; i++) {
		if (strncmp(chassis_rules[i].category, category, MAX_CATEGORY_LENGTH) != 0)
			continue;
		for (j = 0; j < chassis_rules[i].count; j++) {
			do_overlay = true;
			if (j == 0) {
				num = get_rule_num(AUTODET_CHASSIS, (void *)&chassis_rules[i].items[j]);
				if (num == 0) {
					UBOOT_ERROR("failed to get number of keys on %s\n", chassis_rules[i].category);
					return AUTODET_NO_DATA;
				}
				for (k = 0; k < (num-1); k++) {
					// get product config value with key array.
					if (get_product_config_value(chassis_rules[i].items[j].key[k], pcd_value[k]) != 0) {
						UBOOT_ERROR("failed to get key data.(%s)\n", chassis_rules[i].items[j].key[k]);
						return AUTODET_NO_DATA;
					}
				}
			}
			// compare rule data and pcd data.
			for (k = 0; k < (num-1); k++) {
				if (strncmp(chassis_rules[i].items[j].value[k], pcd_value[k], MAX_VALUE_LENGTH) != 0) {
					do_overlay = false;
					break;
				}
			}
			if (do_overlay)
				return sony_dtb_overlay(chassis_rules[i].items[j].value[num-1]);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_pcd_i2c_overlay(const char *first_key, int bus, int slave, int addr, int len)
{
	int i = 0, j = 0;
	int value = -1;
	char first_value[MAX_VALUE_LENGTH] = { 0 };

	// get product config value from json files
	if (get_product_config_value(first_key, first_value) != 0) {
		UBOOT_ERROR("first key error.(%s)\n", first_key);
		return AUTODET_NO_DATA;
	}

	value = check_i2c_status(bus, slave, addr, len);
	if (value == -1) {
		UBOOT_ERROR("i2c data error.(0x%x:0x%x:0x%x)\n", bus, slave, addr);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < pcd_i2c_idx; i++) {
		if (strncmp(pcd_i2c_rules[i].items[0].first_key, first_key, MAX_KEY_LENGTH) != 0)
			continue;

		for (j = 0; j < pcd_i2c_rules[i].count; j++) {
			if ((strncmp(pcd_i2c_rules[i].items[j].first_value, first_value, MAX_VALUE_LENGTH) == 0) &&
				(pcd_i2c_rules[i].items[j].bus == bus) &&
				(pcd_i2c_rules[i].items[j].slave == slave) &&
				(pcd_i2c_rules[i].items[j].addr == addr) &&
				(pcd_i2c_rules[i].items[j].result == value))
				return sony_dtb_overlay(pcd_i2c_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_pcd_adc_overlay(const char *first_key, int adc_ch)
{
	int i = 0, j = 0;
	int value = -1;
	char first_value[MAX_VALUE_LENGTH] = { 0 };

	// get product config value from json files
	if (get_product_config_value(first_key, first_value) != 0) {
		UBOOT_ERROR("first key error.(%s)\n", first_key);
		return AUTODET_NO_DATA;
	}

	value = check_adc_status(adc_ch);
	if (value == -1) {
		UBOOT_ERROR("adc channel error.(0x%x)\n", adc_ch);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < pcd_adc_idx; i++) {
		if (strncmp(pcd_adc_rules[i].items[0].first_key, first_key, MAX_KEY_LENGTH) != 0)
			continue;

		for (j = 0; j < pcd_adc_rules[i].count; j++) {
			if ((strncmp(pcd_adc_rules[i].items[j].first_value, first_value, MAX_VALUE_LENGTH) == 0) &&
				(pcd_adc_rules[i].items[j].channel == adc_ch) &&
				(pcd_adc_rules[i].items[j].min <= value) &&
				(pcd_adc_rules[i].items[j].max >= value))
				return sony_dtb_overlay(pcd_adc_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_pcd_gpio_overlay(const char *first_key, int gpio)
{
	int i = 0, j = 0;
	int value = -1;
	char first_value[MAX_VALUE_LENGTH] = { 0 };

	// get product config value from json files
	if (get_product_config_value(first_key, first_value) != 0) {
		UBOOT_ERROR("first key error.(%s)\n", first_key);
		return AUTODET_NO_DATA;
	}

	value = check_gpio_status(gpio);
	if (value == -1) {
		UBOOT_ERROR("gpio error.(0x%x)\n", gpio);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < pcd_gpio_idx; i++) {
		if (strncmp(pcd_gpio_rules[i].items[0].first_key, first_key, MAX_KEY_LENGTH) != 0)
			continue;

		for (j = 0; j < pcd_gpio_rules[i].count; j++) {
			if ((strncmp(pcd_gpio_rules[i].items[j].first_value, first_value, MAX_VALUE_LENGTH) == 0) &&
				(pcd_gpio_rules[i].items[j].result == value))
				return sony_dtb_overlay(pcd_gpio_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_adc_overlay(int adc_ch)
{
	int i = 0, j = 0;
	int value = -1;

	value = check_adc_status(adc_ch);
	if (value == -1) {
		UBOOT_ERROR("adc error.(0x%x)\n", adc_ch);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < uboot_adc_idx; i++) {
		for (j = 0; j < uboot_adc_rules[i].count; j++) {
			if ((uboot_adc_rules[i].items[j].channel == adc_ch) &&
				(uboot_adc_rules[i].items[j].min <= value) &&
				(uboot_adc_rules[i].items[j].max >= value))
				return sony_dtb_overlay(uboot_adc_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_adc_i2c_overlay(int adc_ch, int bus, int slave, int addr, int len)
{
	int i = 0, j = 0;
	int adc_value = -1, i2c_value = -1;

	adc_value = check_adc_status(adc_ch);
	if (adc_value == -1) {
		UBOOT_ERROR("adc channel error.(0x%x)\n", adc_ch);
		return AUTODET_NO_DATA;
	}

	i2c_value = check_i2c_status(bus, slave, addr, len);
	if (i2c_value == -1) {
		UBOOT_ERROR("i2c data error.(0x%x:0x%x:0x%x)\n", bus, slave, addr);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < adc_i2c_idx; i++) {
		for (j = 0; j < adc_i2c_rules[i].count; j++) {
			if ((adc_i2c_rules[i].items[j].channel == adc_ch) &&
				(adc_i2c_rules[i].items[j].min <= adc_value) &&
				(adc_i2c_rules[i].items[j].max >= adc_value) &&
				(adc_i2c_rules[i].items[j].bus == bus) &&
				(adc_i2c_rules[i].items[j].slave == slave) &&
				(adc_i2c_rules[i].items[j].addr == addr) &&
				(adc_i2c_rules[i].items[j].result == i2c_value))
				return sony_dtb_overlay(adc_i2c_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_gpio_i2c_overlay(int gpio, int bus, int slave, int addr, int len)
{
	int i = 0, j = 0;
	int gpio_value = -1, i2c_value = -1;

	gpio_value = check_gpio_status(gpio);
	if (gpio_value == -1) {
		UBOOT_ERROR("gpio error.(0x%x)\n", gpio);
		return AUTODET_NO_DATA;
	}

	i2c_value = check_i2c_status(bus, slave, addr, len);
	if (i2c_value == -1) {
		UBOOT_ERROR("i2c data error.(0x%x:0x%x:0x%x)\n", bus, slave, addr);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < gpio_i2c_idx; i++) {
		for (j = 0; j < gpio_i2c_rules[i].count; j++) {
			if ((gpio_i2c_rules[i].items[j].gpio == gpio) &&
				(gpio_i2c_rules[i].items[j].result == gpio_value) &&
				(gpio_i2c_rules[i].items[j].bus == bus) &&
				(gpio_i2c_rules[i].items[j].slave == slave) &&
				(gpio_i2c_rules[i].items[j].addr == addr) &&
				(gpio_i2c_rules[i].items[j].value == i2c_value))
				return sony_dtb_overlay(gpio_i2c_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_gpio_overlay(int gpio)
{
	int i = 0, j = 0;
	int value = -1;

	value = check_gpio_status(gpio);
	if (value == -1) {
		UBOOT_ERROR("gpio error.(0x%x)\n", gpio);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < gpio_idx; i++) {
		for (j = 0; j < gpio_rules[i].count; j++) {
			if ((gpio_rules[i].items[j].gpio == gpio) &&
				(gpio_rules[i].items[j].result == value))
				return sony_dtb_overlay(gpio_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_i2c_overlay(int bus, int slave, int addr, int len)
{
	int i = 0, j = 0;
	int value = -1;

	value = check_i2c_status(bus, slave, addr, len);
	if (value == -1) {
		UBOOT_ERROR("i2c error.(0x%x:0x%x:0x%x)\n", bus, slave, addr);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < i2c_idx; i++) {
		for (j = 0; j < i2c_rules[i].count; j++) {
			if ((i2c_rules[i].items[j].bus == bus) &&
				(i2c_rules[i].items[j].slave == slave) &&
				(i2c_rules[i].items[j].addr == addr) &&
				(i2c_rules[i].items[j].result == value))
				return sony_dtb_overlay(i2c_rules[i].items[j].device_name);
		}
	}
	return AUTODET_NO_RULES;
}

int select_sony_panel_overlay(int bus, int slave, int addr)
{
	int i = 0, j = 0;
	uint8_t panel_id[MAX_PANELID_LENGTH + 1] = { 0 };
	unsigned char *cp;

	if (check_panel_status(bus, slave, addr, panel_id) == -1) {
		UBOOT_ERROR("panel error.(0x%x:0x%x:0x%x)\n", bus, slave, addr);
		return AUTODET_NO_DATA;
	}

	for (i = 0; i < panel_idx; i++) {
		for (j = 0; j < panel_rules[i].count; j++) {
			if ((panel_rules[i].items[j].bus == bus) &&
				(panel_rules[i].items[j].slave == slave) &&
				(panel_rules[i].items[j].addr == addr) &&
				(memcmp(panel_rules[i].items[j].panel_id, panel_id, MAX_PANELID_LENGTH) == 0))
				return sony_dtb_overlay(panel_rules[i].items[j].device_name);
		}
	}

	/* panel device debug log */
	printf("----------------------[panel device info start]----------------------\n");
	printf("panel device info=0x%02x:0x%02x:0x%02x\n", bus, slave, addr);
	printf("panel_id:");
	cp = panel_id;
	for (i = 0; i < MAX_PANELID_LENGTH; i++)
		printf(" %02x", *cp++);
	printf("    ");
	cp = panel_id;
	for (i = 0; i < MAX_PANELID_LENGTH; i++) {
		if ((*cp < 0x20) || (*cp > 0x7e))
			printf(".");
		else
			printf("%c", *cp);
		cp++;
	}
	printf("\n");
	printf("----------------------[panel device info end]----------------------\n");

	printf("----------------------[panel_rules info start]----------------------\n");
	for (i = 0; i < panel_idx; i++) {
		for (j = 0; j < panel_rules[i].count; j++)
			printf("panel_rules[%d][%d]=0x%02x:0x%02x:0x%02x:%s:%s\n", i, j,
			panel_rules[i].items[j].bus, panel_rules[i].items[j].slave, panel_rules[i].items[j].addr,
			(char *)panel_rules[i].items[j].panel_id, panel_rules[i].items[j].device_name);
	}
	printf("----------------------[panel_rules info end]----------------------\n");

	return AUTODET_NO_RULES;
}

static int setprop_device_name_for_pmu(const char (*device_name)[MAX_DEVICE_NAME_LENGTH], int count)
{
	int ret = -1, i = 0, length = 0, offset = -1;
	unsigned long dtb_addr = 0;
	char *dtbo_path = NULL, *new_dtbo_path = NULL;
	char fullpath[MAX_VALUE_LENGTH] = { 0 };

	get_uboot_dtb_addr(&dtb_addr);
	offset = fdt_node_offset_by_compatible((const void *)dtb_addr, -1, SONY_CM4_COMPATIBLE);
	if (offset < 0) {
		UBOOT_ERROR("%s node doesn't exist.\n", SONY_CM4_COMPATIBLE);
		return -1;
	}
	dtbo_path = (char *)fdt_getprop((const void *)dtb_addr, offset, SONY_CM4_DTBO_PATH, &length);
	if (!dtbo_path) {
		UBOOT_ERROR("%s property doesn't exist.\n", SONY_CM4_DTBO_PATH);
		return -1;
	}
	new_dtbo_path = malloc(4096);
	if (new_dtbo_path == NULL) {
		UBOOT_ERROR("malloc error.\n");
		return -1;
	}
	if (length > 4096) {
		UBOOT_ERROR("buffer size over.\n");
		free(new_dtbo_path);
		return -1;
	}
	memset(new_dtbo_path, 0, 4096);
	memcpy(new_dtbo_path, dtbo_path, length);
	for (i = 0; i < count; i++) {
		memset(fullpath, 0, sizeof(fullpath));
		snprintf(fullpath, sizeof(fullpath) - 1, "/pmu/%s.dtbo", device_name[i]);
		memcpy(new_dtbo_path + length, fullpath, strlen(fullpath));
		length += (strlen(fullpath) + 1);
	}
	ret = fdt_setprop((void *)dtb_addr, offset, SONY_CM4_DTBO_PATH, (const void *)new_dtbo_path, length);
	free(new_dtbo_path);
	if (ret < 0) {
		UBOOT_ERROR("%s setprop error.\n", SONY_CM4_DTBO_PATH);
		return -1;
	}
	return 0;
}

int select_sony_pcd_overlay_for_pmu(void)
{
	int ret = 0, i = 0, j = 0, k = 0;
	char first_value[MAX_VALUE_LENGTH] = { 0 };
	char second_value[MAX_VALUE_LENGTH] = { 0 };
	char (*device_name)[MAX_DEVICE_NAME_LENGTH] = { 0 };

	device_name = malloc(sizeof(char) * MAX_DEVICE_NAME_LENGTH * pcd_idx);
	if (device_name == NULL) {
		UBOOT_ERROR("malloc error.\n");
		return AUTODET_NO_DATA;
	}
	memset(device_name, 0, sizeof(char) * MAX_DEVICE_NAME_LENGTH * pcd_idx);
	memcpy(device_name[k++], SONY_CM4_DEFAULT_DTB_FILE, strlen(SONY_CM4_DEFAULT_DTB_FILE));
	for (i = 0; i < pcd_idx; i++) {
		// in case of PMU, autodet should execute on u-boot.
		if (strstr(pcd_rules[i].category, SONY_FOR_PMU_SUFFIX) == NULL)
			continue;

		// get product config value with first_key from json files
		memset(first_value, 0, sizeof(first_value));
		if (get_product_config_value(pcd_rules[i].items[0].key[0], first_value) != 0) {
			UBOOT_DEBUG("first key error.(%s)\n", pcd_rules[i].items[0].key[0]);
			continue;
		}
		if (pcd_rules[i].is_multi_key) {
			// get product config value with second_key from json files
			memset(second_value, 0, sizeof(second_value));
			if (get_product_config_value(pcd_rules[i].items[0].key[1], second_value) != 0) {
				UBOOT_DEBUG("second key error.(%s)\n", pcd_rules[i].items[0].key[1]);
				continue;
			}
			for (j = 0; j < pcd_rules[i].count; j++) {
				if ((strncmp(pcd_rules[i].items[j].value[0], first_value, MAX_VALUE_LENGTH) == 0) &&
					(strncmp(pcd_rules[i].items[j].value[1], second_value, MAX_VALUE_LENGTH) == 0)) {
					memcpy(device_name[k++], pcd_rules[i].items[j].value[2], MAX_DEVICE_NAME_LENGTH);
					break;
				}
			}
		} else {
			for (j = 0; j < pcd_rules[i].count; j++) {
				if (strncmp(pcd_rules[i].items[j].value[0], first_value, MAX_VALUE_LENGTH) == 0) {
					memcpy(device_name[k++], pcd_rules[i].items[j].value[1], MAX_DEVICE_NAME_LENGTH);
					break;
				}
			}
		}
	}
	if (k > 0)
		ret = setprop_device_name_for_pmu(device_name, k);

	free(device_name);
	if (ret < 0) {
		UBOOT_ERROR("pmu autodet failed.\n");
		return AUTODET_NO_RULES;
	}
	return 0;
}
