/*
 * Copyright 2022 Sony Corporation
 */

#include <stdio.h>
#include <string.h>
#include <linux/types.h>
#include <debug_impl.h>

#include "data_variation_resolver.h"
#include "data_variation_resolver_client.h"
#include "cJSON.h"

#define DATA_GET 0
#define DATA_NO 1
#define DATA_ERR -1

#define REQUIRED_ATTRIBUTES_KEYS "required_attributes_keys"
#define DATA_VARIATIONS "data_variations"
#define ATTRIBUTES "attributes"
#define DATA_NAME "data_name"
#define ATTRIBUTE_DEFAULT_VALUES "attribute_default_values"
#define ERROR_STR ""

static int32_t compare_attributes_json_string(const cJSON *variations, char *attibutes, int32_t attibutes_size);
static int32_t search_attributes(const cJSON *attributes_array, int32_t attibutes_size, const char *attibutes);
static int32_t get_attributes_default_value(const cJSON *json, int32_t attibutes_size, char *attibutes);
static void get_data_name(const cJSON *variations, int32_t data_name_size, char *data_name);

void get_require_attribute_keys(const char *json_string, int32_t list_size, char *list)
{

	cJSON *json;
	cJSON *keys_array;
	cJSON *key;
	uint32_t keys_array_size;
	uint32_t key_strlen;
	char key_store[KEY_MAXSIZE][CHAR_MAXSIZE] = {0};

	//list NULL check
	if (list == NULL) {
		UBOOT_ERROR("list NULL\n");
		return;
	}
	memset(list, 0, list_size);

	//json_string NULL check
	if (json_string == NULL) {
		UBOOT_ERROR("json_string NULL\n");
		return;
	}

	if (list_size != sizeof(key_store)) {
		UBOOT_ERROR("list_size and key_store size not match\n");
		return;
	}

	//JSON Parse process
	json = cJSON_Parse(json_string);
	if (json == NULL) {
		UBOOT_ERROR("%s cJSON_Parse Error\n", __func__);
		cJSON_Delete(json);
		return;
	}

	//Get the value of "required_attributes_keys"
	keys_array = cJSON_GetObjectItem(json, REQUIRED_ATTRIBUTES_KEYS);
	if (keys_array == NULL) {
		UBOOT_ERROR("required_attributes_keys cJSON_GetObjectItem Error\n");
		cJSON_Delete(json);
		return;
	}

	//Get the number of value of "required_attributes_keys"
	keys_array_size = cJSON_GetArraySize(keys_array);
	if (keys_array_size == 0) {
		UBOOT_ERROR("keys_array_size cJSON_GetArraySize Error\n");
		cJSON_Delete(json);
		return;
	}

	if (keys_array_size > KEY_MAXSIZE) {
		UBOOT_ERROR("KEY_MAXSIZE Error\n");
		cJSON_Delete(json);
		return;
	}

	///Put Value in the list
	for (uint32_t i = 0; i < keys_array_size; i++) {
		key = cJSON_GetArrayItem(keys_array, i);
		if (key == NULL) {
			UBOOT_ERROR("key cJSON_GetArrayItem Error\n");
			cJSON_Delete(json);
			return;
		}
		key_strlen = strnlen(key->valuestring, CHAR_MAXSIZE);
		strncpy(key_store[i], key->valuestring, key_strlen);
	}
	memcpy(list, key_store, list_size);
	cJSON_Delete(json);
}

void get_corresponding_data_variation(const char *json_string, int32_t attibutes_size, const char *attibutes,  int32_t data_name_size, char *data_name)
{

	cJSON *json;
	cJSON *variations_array;
	cJSON *variations;
	cJSON *keys_array;
	uint32_t variations_array_size;
	uint32_t keys_array_size;
	uint32_t list_keynum = 0;
	int32_t ret = DATA_ERR;
	uint32_t array_num;
	char attibutes_copy[KEY_VALUE_NUM][KEY_MAXSIZE][CHAR_MAXSIZE] = {0};

	//data_name NULL check
	if (data_name == NULL) {
		UBOOT_ERROR("data_name NULL\n");
		return;
	}
	memset(data_name, 0, data_name_size);

	//json_string NULL check
	if (json_string == NULL) {
		UBOOT_ERROR("json_string NULL\n");
		return;
	}

	//attibutes NULL check
	if (attibutes == NULL) {
		UBOOT_ERROR("attibutes NULL\n");
		return;
	}

	if (attibutes_size != sizeof(attibutes_copy)) {
		UBOOT_ERROR("attibutes_size and attibutes_copy size not match\n");
		return;
	}
	memcpy(attibutes_copy, attibutes, attibutes_size);

	//JSON Parse process
	json = cJSON_Parse(json_string);
	if (json == NULL) {
		UBOOT_ERROR("%s cJSON_Parse Error\n", __func__);
		cJSON_Delete(json);
		return;
	}

	//Get the value of "required_attributes_keys"
	keys_array = cJSON_GetObjectItem(json, REQUIRED_ATTRIBUTES_KEYS);
	if (keys_array == NULL) {
		UBOOT_ERROR("required_attributes_keys cJSON_GetObjectItem Error\n");
		cJSON_Delete(json);
		return;
	}

	//Get the number of value of "required_attributes_keys"
	keys_array_size = cJSON_GetArraySize(keys_array);
	if (keys_array_size == 0) {
		UBOOT_ERROR("keys_array_size cJSON_GetArraySize Error\n");
		cJSON_Delete(json);
		return;
	}

	//Compare the number of unordered_map Keys with the number of "required_attributes_keys"
	for (uint32_t i = 0; i < KEY_MAXSIZE; i++) {
		if (strnlen(attibutes_copy[KEY_IDX][i], CHAR_MAXSIZE) == 0)
			break;
		list_keynum++;
	}
	if (keys_array_size != list_keynum) {
		UBOOT_ERROR("key num do not match list\n");
		cJSON_Delete(json);
		return;
	}

	//Get the value of "data_variations"
	variations_array = cJSON_GetObjectItem(json, DATA_VARIATIONS);
	if (variations_array == NULL) {
		UBOOT_ERROR("data_variations cJSON_GetObjectItem Error\n");
		cJSON_Delete(json);
		return;
	}

	//Get the number of value of "recipe_data_variations"
	variations_array_size = cJSON_GetArraySize(variations_array);
	if (variations_array_size == 0) {
		UBOOT_ERROR("variations_array_size cJSON_GetArraySize Error\n");
		cJSON_Delete(json);
		return;
	}

	//Get default value
	ret = get_attributes_default_value(json, attibutes_size, &attibutes_copy[0][0][0]);
	if (ret == DATA_ERR) {
		cJSON_Delete(json);
		return;
	}

	//Comparison of argument attibutes and json_string
	for (array_num = 0; array_num < variations_array_size; array_num++) {
		variations = cJSON_GetArrayItem(variations_array, array_num);
		if (variations == NULL) {
			UBOOT_ERROR("variations cJSON_GetArrayItem Error\n");
			ret = DATA_ERR;
			break;
		}
		ret = compare_attributes_json_string(variations, &attibutes_copy[0][0][0], attibutes_size);
		if (ret != DATA_NO)
			break;
	}

	//Get data_name
	if (ret == DATA_GET)
		get_data_name(variations, data_name_size, data_name);
	else if (array_num == variations_array_size)
		UBOOT_ERROR("data not found\n");

	cJSON_Delete(json);
}

static int32_t compare_attributes_json_string(const cJSON *variations, char *attibutes, int32_t attibutes_size)
{
	cJSON *attributes_array;
	int32_t ret = DATA_ERR;

	if (attibutes == NULL) {
		UBOOT_ERROR("attibutes NULL\n");
		return ret;
	}
	if (variations == NULL) {
		UBOOT_ERROR("variations NULL\n");
		return ret;
	}

	//Get json_string attributes
	attributes_array = cJSON_GetObjectItem(variations, ATTRIBUTES);
	if (attributes_array == NULL) {
		UBOOT_ERROR("attributes_array cJSON_GetObjectItem Error\n");
		return ret;
	}

	//Search data_name
	ret = search_attributes(attributes_array, attibutes_size, attibutes);

	return ret;
}

static int32_t search_attributes(const cJSON *attributes_array, int32_t attibutes_size, const char *attibutes)
{
	int32_t ret = DATA_ERR;
	int32_t attributes_array_size;
	cJSON *attributes;
	cJSON *key;
	char attibutes_copy[KEY_VALUE_NUM][KEY_MAXSIZE][CHAR_MAXSIZE] = {0};

	if (attributes_array == NULL) {
		UBOOT_ERROR("attributes_array NULL\n");
		return ret;
	}
	if (attibutes == NULL) {
		UBOOT_ERROR("attibutes NULL\n");
		return ret;
	}

	if (attibutes_size != sizeof(attibutes_copy)) {
		UBOOT_ERROR("attibutes_size and attibutes_copy size not match\n");
		return ret;
	}
	memcpy(attibutes_copy, attibutes, attibutes_size);

	//Get the number of contents of json_string attributes
	attributes_array_size = cJSON_GetArraySize(attributes_array);
	if (attributes_array_size == 0) {
		UBOOT_ERROR("attributes_array_size cJSON_GetArraySize Error\n");
		return ret;
	}

	for (uint32_t j = 0; j < attributes_array_size; j++) {
		//Get the value of contents of json_string attributes
		attributes = cJSON_GetArrayItem(attributes_array, j);
		if (attributes == NULL) {
			UBOOT_ERROR("attributes cJSON_GetArrayItem Error\n");
			ret = DATA_ERR;
			break;
		}

		//Compare with the contents of argument attibutes
		for (uint32_t k = 0; k < KEY_MAXSIZE; k++) {
			if (strcmp(attibutes_copy[KEY_IDX][k], ERROR_STR) == 0)
				break;
			key = cJSON_GetObjectItem(attributes, attibutes_copy[KEY_IDX][k]);
			if (key == NULL) {
				UBOOT_ERROR("%s cJSON_GetObjectItem Error\n", attibutes_copy[KEY_IDX][k]);
				ret = DATA_ERR;
				break;
			}
			if (strcmp(attibutes_copy[VALUE_IDX][k], key->valuestring) != 0) {
				ret = DATA_NO;
				break;
			}
			ret = DATA_GET;
		}
		if ((ret == DATA_ERR) || (ret == DATA_GET))
			break;
	}
	return ret;
}

static int32_t get_attributes_default_value(const cJSON *json, int32_t attibutes_size, char *attibutes)
{
	int32_t ret = DATA_ERR;
	int32_t default_key_len;
	cJSON *default_key;
	cJSON *attribute_default_values;
	char attibutes_copy[KEY_VALUE_NUM][KEY_MAXSIZE][CHAR_MAXSIZE] = {0};

	if (json == NULL) {
		UBOOT_ERROR("json NULL\n");
		return ret;
	}
	if (attibutes == NULL) {
		UBOOT_ERROR("attibutes NULL\n");
		return ret;
	}
	if (attibutes_size != sizeof(attibutes_copy)) {
		UBOOT_ERROR("attibutes_size and attibutes_copy size not match\n");
		return ret;
	}
	memcpy(attibutes_copy, attibutes, attibutes_size);

	//Get the value of "attribute_default_values"
	attribute_default_values = cJSON_GetObjectItem(json, ATTRIBUTE_DEFAULT_VALUES);

	for (uint32_t k = 0; k < KEY_MAXSIZE; k++) {
		if (strcmp(attibutes_copy[KEY_IDX][k], ERROR_STR) == 0)
			break;
		if (strcmp(attibutes_copy[VALUE_IDX][k], ERROR_STR) == 0) {
			//Get default value
			if (attribute_default_values == NULL) {
				UBOOT_ERROR("attribute_default_values NULL\n");
				return ret;
			}
			default_key = cJSON_GetObjectItem(attribute_default_values, attibutes_copy[KEY_IDX][k]);
			if (default_key == NULL) {
				UBOOT_ERROR("%s Get default value Error\n", attibutes_copy[KEY_IDX][k]);
				return ret;
			}
			default_key_len = strlen(default_key->valuestring);
			if (default_key_len > CHAR_MAXSIZE) {
				UBOOT_ERROR("default_key size Error\n");
				return ret;
			}
			strncpy(attibutes_copy[VALUE_IDX][k], default_key->valuestring, default_key_len);
			UBOOT_INFO("use Default Value KEY : %s, VALUE : %s\n", attibutes_copy[KEY_IDX][k], attibutes_copy[VALUE_IDX][k]);
		}
	}

	memcpy(attibutes, attibutes_copy, attibutes_size);
	ret = DATA_GET;

	return ret;
}

static void get_data_name(const cJSON *variations, int32_t data_name_size, char *data_name)
{
	cJSON *data;
	uint32_t data_namelen;

	if (variations == NULL) {
		UBOOT_ERROR("variations NULL\n");
		return;
	}
	if (data_name == NULL) {
		UBOOT_ERROR("data_name NULL\n");
		return;
	}

	data = cJSON_GetObjectItem(variations, DATA_NAME);
	if (data == NULL) {
		UBOOT_ERROR("data_name cJSON_GetObjectItem Error\n");
	} else {
		data_namelen = strnlen(data->valuestring, CHAR_MAXSIZE);
		if (data_namelen > data_name_size)
			UBOOT_ERROR("data_name_size Error\n");
		else
			strncpy(data_name, data->valuestring, data_namelen);
	}
}
