/*
 * Copyright 2021 Sony Corporation
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <linux/types.h>
#include <debug_impl.h>
#include <include/sony_fs.h>
#include <adc.h>
#include <dm/uclass.h>
#include <prodconfig.h>
#include <utility.h>

#include "vbo_frame_info.h"

extern long strtol(const char *cp, char **endp, unsigned int base);

#define KEY_LENGTH 4
#define KEY_WORD "key"
#define KEY_WORD_LENGTH 3
#define IKEY_LENGTH 4
#define IKEY_WORD "ikey"
#define IKEY_WORD_LENGTH 4
#define TID_LENGTH 4
#define PACKET_WORD "packets"
#define PACKET_WORD_LENGTH 7
#define HEXADECIMAL 16
#define DECIMAL 10
#define BOOT_STATE_NUM_MAX 10
#define DET_8K_CONNECTED_VAL 488

#define ADC_CH_PANEL_DET_BE_DET 4

#define HAS_8KBE_KEY "8KBE"
#define NO_8KBE_VALUE "-"
#define HAS_8KBE 1
#define NO_8KBE 0
#define HAS_8KBE_WORD_LENGTH MAX_PRODUCT_CONFIG_LENGTH

#define BCM_UPDATE_DISABLE_FLAG "0"
#define BCM_UPDATE_FLAG_SIZE 1
#define BCM_UPDATE_DISABLE 0
#define BCM_UPDATE_ENABLE 1
#define UENV_PART_NAME "uenv"
#define BCM_STATUS_FILE_PATH "/fwupdate/bcm_status.txt"

// csv file path
static const char FRAMEINFO_PATH[] = "/pq/be_boot/frame_info.csv";

struct ST_PACKETS {
	int32_t boot_state_num;
	int32_t frame_header_size;
	int32_t frame_num_in_pkt;
	char    tid[TID_NUM_SIZE_MAX][STR_LENGTH];
};

struct ST_FRAME_INFO {
	char       key[STR_LENGTH];
	char       ikey[STR_LENGTH];
	int32_t    send_frame_num[BOOT_STATE_NUM_MAX];
	struct ST_PACKETS packets[FRAME_HEADER_NUM_MAX];
};

struct ST_FRAME_INFO_PANEL {
	uint8_t boot_mode;
	uint8_t panel_off_boot_request;
	uint8_t logo_luminance_limit;
	uint8_t has_8kbe;
	uint8_t is_8kbe_connected;
	uint8_t bcm_firmware_update_request;
	uint8_t firmware_update_request_8k;
	uint32_t panel_on_time;
};

static uint16_t crc16tab[256] = {
	0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
	0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
	0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
	0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
	0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
	0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
	0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
	0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
	0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
	0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
	0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
	0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
	0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
	0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
	0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
	0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
	0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
	0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
	0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
	0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
	0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
	0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
	0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
	0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
	0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
	0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
	0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
	0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
	0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
	0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
	0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
	0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
};

static enum E_VBO_FRAME_STATUS load_vb1_frame_info(struct ST_FRAME_INFO *frame_info);
static enum E_VBO_FRAME_STATUS get_key_and_ikey(char *csv_file, uint32_t csv_size, struct ST_FRAME_INFO *frame_info);
static enum E_VBO_FRAME_STATUS get_packets(char *csv_file, uint32_t csv_size, struct ST_FRAME_INFO *frame_info);
static enum E_VBO_FRAME_STATUS make_vb1_frame_info_panel(struct ST_FRAME_INFO_PANEL *frame_info_panel);
static enum E_VBO_FRAME_STATUS get_has_8kbe(uint8_t *has_8kbe);
static enum E_VBO_FRAME_STATUS get_adc_det(unsigned int *adc_data);
static enum E_VBO_FRAME_STATUS get_panel_off_boot_request(uint8_t *is_panel_off_boot_request);
static enum E_VBO_FRAME_STATUS get_bcm_fwupdate_status(uint8_t *is_bcm_fwupate_update_request);

enum E_VBO_FRAME_STATUS make_vb1_frame_header(struct ST_FRAME_HEADER *frame_header)
{
	struct ST_FRAME_INFO frame_info;
	struct ST_FRAME_INFO_PANEL frame_info_panel;
	enum E_VBO_FRAME_STATUS error_num;
	uint32_t boot_state_mem = 0;
	uint32_t frame_number_count = 1;
	uint32_t boot_state_index = 0;
	uint32_t total_frame_num = 0;

	if (frame_header == NULL) {
		UBOOT_ERROR("frame_header NULL\n");
		return E_VBO_FRAME_POINTER_NULL;
	}

	memset(&frame_info, 0, sizeof(frame_info));
	memset(&frame_info_panel, 0, sizeof(frame_info_panel));
	memset(&frame_header->param, 0, sizeof(frame_header->param));

	error_num = load_vb1_frame_info(&frame_info);
	if (error_num != 0)
		return error_num;

	error_num = make_vb1_frame_info_panel(&frame_info_panel);
	if (error_num != 0)
		return error_num;

	for (int i = 0; i < frame_header->num; i++) {

		if (boot_state_mem != frame_info.packets[i].boot_state_num) {
			boot_state_mem = frame_info.packets[i].boot_state_num;
			total_frame_num = frame_info.send_frame_num[boot_state_index];
			frame_number_count = 1;
			boot_state_index++;
		}

		// key
		frame_header->param[i][0]  = frame_info.key[0];
		frame_header->param[i][1]  = frame_info.key[1];
		frame_header->param[i][2]  = frame_info.key[2];
		frame_header->param[i][3]  = frame_info.key[3];
		// Frame Header
		frame_header->param[i][4]  = (frame_info.packets[i].frame_header_size >> 24) & (0xFF);
		frame_header->param[i][5]  = (frame_info.packets[i].frame_header_size >> 16) & (0xFF);
		frame_header->param[i][6]  = (frame_info.packets[i].frame_header_size >> 8)  & (0xFF);
		frame_header->param[i][7]  = (frame_info.packets[i].frame_header_size)       & (0xFF);
		// Frame Size(will be set later)
		frame_header->param[i][8]  = 0;
		frame_header->param[i][9]  = 0;
		frame_header->param[i][10] = 0;
		frame_header->param[i][11] = 0;
		// BE boot State
		frame_header->param[i][12] = (frame_info.packets[i].boot_state_num >> 24) & (0xFF);
		frame_header->param[i][13] = (frame_info.packets[i].boot_state_num >> 16) & (0xFF);
		frame_header->param[i][14] = (frame_info.packets[i].boot_state_num >> 8)  & (0xFF);
		frame_header->param[i][15] = (frame_info.packets[i].boot_state_num)       & (0xFF);
		// Frame number
		frame_header->param[i][16] = (frame_number_count >> 24) & (0xFF);
		frame_header->param[i][17] = (frame_number_count >> 16) & (0xFF);
		frame_header->param[i][18] = (frame_number_count >> 8)  & (0xFF);
		frame_header->param[i][19] = (frame_number_count)       & (0xFF);
		// Send Frame number
		frame_header->param[i][20] = (total_frame_num >> 24) & (0xFF);
		frame_header->param[i][21] = (total_frame_num >> 16) & (0xFF);
		frame_header->param[i][22] = (total_frame_num >> 8)  & (0xFF);
		frame_header->param[i][23] = (total_frame_num)       & (0xFF);
		// Amount in Frame
		frame_header->param[i][24] = (frame_info.packets[i].frame_num_in_pkt >> 24) & (0xFF);
		frame_header->param[i][25] = (frame_info.packets[i].frame_num_in_pkt >> 16) & (0xFF);
		frame_header->param[i][26] = (frame_info.packets[i].frame_num_in_pkt >> 8)  & (0xFF);
		frame_header->param[i][27] = (frame_info.packets[i].frame_num_in_pkt)       & (0xFF);
		// Reserve
		frame_header->param[i][28] = 0;
		frame_header->param[i][29] = 0;
		// HeaderCRC(will be set later)
		frame_header->param[i][30] = 0;
		frame_header->param[i][31] = 0;

		for (int k = 0; k < frame_info.packets[i].frame_num_in_pkt; k++)
			strncpy(frame_header->packets_frame[i].tid_frame[k], frame_info.packets[i].tid[k], sizeof(frame_header->packets_frame[i].tid_frame[k])-1);
		if (boot_state_mem >= 2) {
			// BootMode(0:Cold, 1: Hot)
			frame_header->param[i][32] = frame_info_panel.boot_mode;
			// Panel Off Boot Request
			frame_header->param[i][33] = frame_info_panel.panel_off_boot_request;
			//Logo Luminance Limit
			frame_header->param[i][34] = frame_info_panel.logo_luminance_limit;
			//has_8kbe
			frame_header->param[i][35] = frame_info_panel.has_8kbe;
			//is_8kbe_connected
			frame_header->param[i][36] = frame_info_panel.is_8kbe_connected;
			//BCM Firmware Update Request
			frame_header->param[i][37] = frame_info_panel.bcm_firmware_update_request;
			//8KBE Firmware Update Request
			frame_header->param[i][38] = frame_info_panel.firmware_update_request_8k;
			//panel_on_time[us]
			frame_header->param[i][40] = (frame_info_panel.panel_on_time >> 24) & (0xFF);
			frame_header->param[i][41] = (frame_info_panel.panel_on_time >> 16) & (0xFF);
			frame_header->param[i][42] = (frame_info_panel.panel_on_time >> 8)  & (0xFF);
			frame_header->param[i][43] = (frame_info_panel.panel_on_time)       & (0xFF);
		}
		frame_number_count++;
	}
	frame_header->ikey[0] = frame_info.ikey[0];
	frame_header->ikey[1] = frame_info.ikey[1];
	frame_header->ikey[2] = frame_info.ikey[2];
	frame_header->ikey[3] = frame_info.ikey[3];

	return E_VBO_FRAME_OK;
}

uint16_t calc_vb1_header_crc(uint8_t *buf, uint32_t len)
{
	uint32_t counter;
	uint16_t crc = 0;

	for (counter = 0; counter < len; counter++)
		crc = (crc << 8) ^ crc16tab[((crc >> 8) ^ *(char *)buf++) & 0x00FF];
	return crc;
}

int8_t load_frame_info_num(void)
{
	char *p;
	char *csv_file;
	uint8_t frame_num = 0;
	int32_t csv_size = -1;
	int32_t load_ret = -1;

	csv_size = ext4_f_size_mmc(DATA_PART_NAME, FRAMEINFO_PATH);
	if (csv_size == -1) {
		UBOOT_ERROR("csv file size error.\n");
		return -1;
	}

	csv_file = (char *)malloc(csv_size + 1);
	if (csv_file == NULL) {
		UBOOT_ERROR("data malloc error_csv_file\n");
		return -1;
	}
	memset((void *)csv_file, 0, csv_size + 1);

	load_ret = ext4_f_load_mmc(DATA_PART_NAME, FRAMEINFO_PATH, csv_file);
	if (load_ret == -1) {
		UBOOT_ERROR("frameinfo open failed\n");
		free(csv_file);
		return -1;
	}

	p = strtok(csv_file, "\n");
	while (p != NULL) {
		if (strncmp(p, PACKET_WORD, PACKET_WORD_LENGTH) == 0)
			frame_num++;
		p = strtok(NULL, "\n");
	}
	free(csv_file);
	return frame_num;
}

static enum E_VBO_FRAME_STATUS load_vb1_frame_info(struct ST_FRAME_INFO *frame_info)
{
	char *csv_file;
	int32_t csv_size = -1;
	int32_t load_ret = -1;
	enum E_VBO_FRAME_STATUS ret;

	if (frame_info == NULL) {
		UBOOT_ERROR("frame_info NULL\n");
		return E_VBO_FRAME_POINTER_NULL;
	}

	csv_size = ext4_f_size_mmc(DATA_PART_NAME, FRAMEINFO_PATH);
	if (csv_size == -1) {
		UBOOT_ERROR("csv file size error.\n");
		return E_VBO_FRAME_FILE_SIZE_NG;
	}

	csv_file = (char *)malloc(csv_size + 1);
	if (csv_file == NULL) {
		UBOOT_ERROR("data malloc error_csv_file\n");
		return E_VBO_FRAME_MALLOC_NG;
	}
	memset((void *)csv_file, 0, csv_size + 1);

	load_ret = ext4_f_load_mmc(DATA_PART_NAME, FRAMEINFO_PATH, csv_file);
	if (load_ret == -1) {
		UBOOT_ERROR("frameinfo open failed\n");
		free(csv_file);
		return E_VBO_FRAME_CSV_FILEOPEN_NG;
	}

	for (int i = 0; i < csv_size; i++) {
		if ('\n' == csv_file[i])
			csv_file[i] = ',';
	}

	// get key and ikye
	ret = get_key_and_ikey(csv_file, csv_size, frame_info);
	if (ret != E_VBO_FRAME_OK) {
		free(csv_file);
		return ret;
	}

	// get packets
	ret = get_packets(csv_file, csv_size, frame_info);
	if (ret != E_VBO_FRAME_OK) {
		free(csv_file);
		return ret;
	}

	free(csv_file);
	return E_VBO_FRAME_OK;
}

static enum E_VBO_FRAME_STATUS get_key_and_ikey(char *csv_file, uint32_t csv_size, struct ST_FRAME_INFO *frame_info)
{
	char *key_data;
	char *ikey_data;
	char *csv_file_key;
	char *csv_file_ikey;

	if (frame_info == NULL) {
		UBOOT_ERROR("frame_info NULL\n");
		return E_VBO_FRAME_POINTER_NULL;
	}
	if (csv_file == NULL) {
		UBOOT_ERROR("csv_file NULL\n");
		return E_VBO_FRAME_POINTER_NULL;
	}

	csv_file_key = (char *)malloc(csv_size + 1);
	if (csv_file_key == NULL) {
		UBOOT_ERROR("data malloc error_csv_file_key\n");
		return E_VBO_FRAME_MALLOC_NG;
	}
	memset((void *)csv_file_key, 0, csv_size + 1);
	memcpy(csv_file_key, csv_file, csv_size + 1);

	csv_file_ikey = (char *)malloc(csv_size + 1);
	if (csv_file_ikey == NULL) {
		UBOOT_ERROR("data malloc error_csv_file_ikey\n");
		free(csv_file_key);
		return E_VBO_FRAME_MALLOC_NG;
	}
	memset((void *)csv_file_ikey, 0, csv_size + 1);
	memcpy(csv_file_ikey, csv_file, csv_size + 1);

	// get key
	key_data = strtok(csv_file_key, ",");
	while (key_data != NULL) {
		if (strncmp(key_data, KEY_WORD, KEY_WORD_LENGTH) == 0)
			break;
		key_data = strtok(NULL, ",");
	}
	if (key_data == NULL) {
		UBOOT_ERROR("key format error\n");
		free(csv_file_key);
		free(csv_file_ikey);
		return E_VBO_FRAME_KEY_FORMAT_NG;
	}

	key_data = strtok(NULL, ",");
	if (key_data != NULL) {
		if (strnlen(key_data, STR_LENGTH) == KEY_LENGTH) {
			strncpy(frame_info->key, key_data, KEY_LENGTH);
		} else {
			UBOOT_ERROR("key data format error\n");
			free(csv_file_key);
			free(csv_file_ikey);
			return E_VBO_FRAME_KEY_DATA_NG;
		}
	} else {
		UBOOT_ERROR("key line error\n");
		free(csv_file_key);
		free(csv_file_ikey);
		return E_VBO_FRAME_KEY_LINE_NG;
	}

	// get ikey
	ikey_data = strtok(csv_file_ikey, ",");
	while (ikey_data != NULL) {
		if (strncmp(ikey_data, IKEY_WORD, IKEY_WORD_LENGTH) == 0)
			break;
		ikey_data = strtok(NULL, ",");
	}
	if (ikey_data == NULL) {
		UBOOT_ERROR("ikey format error\n");
		free(csv_file_key);
		free(csv_file_ikey);
		return E_VBO_FRAME_IKEY_FORMAT_NG;
	}

	ikey_data = strtok(NULL, ",");
	if (ikey_data != NULL) {
		if (strnlen(ikey_data, STR_LENGTH) == IKEY_LENGTH) {
			strncpy(frame_info->ikey, ikey_data, IKEY_LENGTH);
		} else {
			UBOOT_ERROR("ikey data format error\n");
			free(csv_file_key);
			free(csv_file_ikey);
			return E_VBO_FRAME_IKEY_DATA_NG;
		}
	} else {
		UBOOT_ERROR("ikey line error\n");
		free(csv_file_key);
		free(csv_file_ikey);
		return E_VBO_FRAME_IKEY_LINE_NG;
	}

	free(csv_file_key);
	free(csv_file_ikey);
	return E_VBO_FRAME_OK;
}

static enum E_VBO_FRAME_STATUS get_packets(char *csv_file, uint32_t csv_size, struct ST_FRAME_INFO *frame_info)
{
	char *strtol_error;
	char *csv_file_packet;
	char *packet_data;
	int32_t tid_num = 0;
	int32_t packet_num = 0;
	int32_t send_frame_count = 0;
	int32_t boot_state_mem = 1;
	int32_t boot_state_index = 0;

	if (frame_info == NULL) {
		UBOOT_ERROR("frame_info NULL\n");
		return E_VBO_FRAME_POINTER_NULL;
	}
	if (csv_file == NULL) {
		UBOOT_ERROR("csv_file NULL\n");
		return E_VBO_FRAME_POINTER_NULL;
	}

	csv_file_packet = (char *)malloc(csv_size + 1);
	if (csv_file_packet == NULL) {
		UBOOT_ERROR("data malloc error_csv_file_copy\n");
		return E_VBO_FRAME_MALLOC_NG;
	}
	memset((void *)csv_file_packet, 0, csv_size + 1);
	memcpy(csv_file_packet, csv_file, csv_size + 1);

	packet_data = strtok(csv_file_packet, ",");
	while (packet_data != NULL) {
		if (strncmp(packet_data, PACKET_WORD, PACKET_WORD_LENGTH) == 0)
			break;
		packet_data = strtok(NULL, ",");
	}
	if (packet_data == NULL) {
		UBOOT_ERROR("packet not found\n");
		free(csv_file_packet);
		return E_VBO_FRAME_NO_PACKET_LINE;
	}

	// get packets
	while (packet_data != NULL) {
		if (strncmp(packet_data, PACKET_WORD, PACKET_WORD_LENGTH) == 0) {
			packet_data = strtok(NULL, ",");
			frame_info->packets[packet_num].boot_state_num = (int)strtol(packet_data, &strtol_error, DECIMAL);
			if (frame_info->packets[packet_num].boot_state_num == 0) {
				UBOOT_ERROR("boot state error\n");
				free(csv_file_packet);
				return E_VBO_FRAME_BOOT_STATE_NG;
			}
			packet_data = strtok(NULL, ",");
			frame_info->packets[packet_num].frame_header_size = (int)strtol(packet_data, &strtol_error, HEXADECIMAL);
			if (frame_info->packets[packet_num].frame_header_size == 0) {
				UBOOT_ERROR("frame header error\n");
				free(csv_file_packet);
				return E_VBO_FRAME_FRAME_HEADER_NG;
			}
			packet_data = strtok(NULL, ",");
			while ((packet_data != NULL) && (strnlen(packet_data, STR_LENGTH) == TID_LENGTH)) {
				strncpy(frame_info->packets[packet_num].tid[tid_num], packet_data, sizeof(frame_info->packets[packet_num].tid[tid_num])-1);
				tid_num++;
				packet_data = strtok(NULL, ",");
			}

			frame_info->packets[packet_num].frame_num_in_pkt = tid_num;
			if (boot_state_mem != frame_info->packets[packet_num].boot_state_num) {
				frame_info->send_frame_num[boot_state_index] = send_frame_count;
				boot_state_mem = frame_info->packets[packet_num].boot_state_num;
				send_frame_count = 0;
				boot_state_index++;
			}
			packet_num++;
			send_frame_count++;
		} else {
			packet_data = strtok(NULL, ",");
		}
		tid_num = 0;
	}
	frame_info->send_frame_num[boot_state_index] = send_frame_count;
	if (packet_num == 0) {
		UBOOT_ERROR("no pkt line\n");
		free(csv_file_packet);
		return E_VBO_FRAME_NO_PACKET_LINE;
	}

	free(csv_file_packet);
	return E_VBO_FRAME_OK;
}

static enum E_VBO_FRAME_STATUS make_vb1_frame_info_panel(struct ST_FRAME_INFO_PANEL *frame_info_panel)
{
	unsigned int adc_data;
	int ret;
	uint8_t has_8Kbe_val;
	uint8_t is_panel_off_boot_request;
	uint8_t is_bcm_fwupate_update_request;

	if (frame_info_panel == NULL) {
		UBOOT_ERROR("frame_info_panel NULL\n");
		return E_VBO_FRAME_POINTER_NULL;
	}

	//BootMode
	frame_info_panel->boot_mode = 0;

	//Panel Off Boot Request
	ret = get_panel_off_boot_request(&is_panel_off_boot_request);
	if (ret != 0)
		return ret;

	frame_info_panel->panel_off_boot_request = is_panel_off_boot_request;

	// Logo Luminance Limit
	if (is_panel_off_boot_request == 0) {
		// others
		frame_info_panel->logo_luminance_limit = 1;
	} else {
		// Panel Off Boot
		frame_info_panel->logo_luminance_limit = 0;
	}

	//has_8kbe
	ret = get_has_8kbe(&has_8Kbe_val);
	if (ret != 0) {
		UBOOT_ERROR("get_has_8kbe Error\n");
		frame_info_panel->has_8kbe = NO_8KBE;
	} else {
		frame_info_panel->has_8kbe = has_8Kbe_val;
	}

	//PANEL_DET/BE_DET
	ret = get_adc_det(&adc_data);
	if (ret != 0)
		return ret;

	//is_8kbe_connected
	if (adc_data <= DET_8K_CONNECTED_VAL)
		frame_info_panel->is_8kbe_connected = 1; //FY22_PQSYS_fixme
	else
		frame_info_panel->is_8kbe_connected = 0; //FY22_PQSYS_fixme

	//BCM Firmware Update Request
	ret = get_bcm_fwupdate_status(&is_bcm_fwupate_update_request);
	if (ret != 0) {
		UBOOT_ERROR("get_bcm_fwupdate_status Error\n");
		frame_info_panel->bcm_firmware_update_request = BCM_UPDATE_ENABLE;
	} else {
		frame_info_panel->bcm_firmware_update_request = is_bcm_fwupate_update_request;
	}

	//8KBE Firmware Update Request
	frame_info_panel->firmware_update_request_8k = 0; //FY24_PQSYS_fixme

	// panel_on_time[us]
	frame_info_panel->panel_on_time = 0; //FY22_PQSYS_fixme

	return E_VBO_FRAME_OK;
}

static enum E_VBO_FRAME_STATUS get_has_8kbe(uint8_t *has_8kbe)
{
	int ret;
	char has_8Kbe_value[HAS_8KBE_WORD_LENGTH] = {0};

	ret = get_product_config_value(HAS_8KBE_KEY, has_8Kbe_value);

	if (ret != 0) {
		UBOOT_ERROR("get_product_config_value Error\n");
		return E_VBO_FRAME_PRODUCT_CONFIG_ERROR;
	}
	if (strcmp(has_8Kbe_value, NO_8KBE_VALUE) == 0)
		*has_8kbe = NO_8KBE;
	else
		*has_8kbe = HAS_8KBE;

	return E_VBO_FRAME_OK;
}

static enum E_VBO_FRAME_STATUS get_adc_det(unsigned int *adc_data)
{
	struct udevice *adc_dev;
	int ret;

	ret = uclass_get_device(UCLASS_ADC, 0, &adc_dev);
	if (ret != 0) {
		UBOOT_ERROR("uclass_get_device error\n");
		return E_VBO_FRAME_ADC_NG;
	}
	ret = adc_start_channel(adc_dev, ADC_CH_PANEL_DET_BE_DET);
	if (ret != 0) {
		UBOOT_ERROR("adc_start_channel error\n");
		return E_VBO_FRAME_ADC_NG;
	}
	ret = adc_channel_data(adc_dev, ADC_CH_PANEL_DET_BE_DET, adc_data);
	if (ret != 0) {
		UBOOT_ERROR("adc_channel_data error\n");
		return E_VBO_FRAME_ADC_NG;
	}
	return E_VBO_FRAME_OK;
}

static enum E_VBO_FRAME_STATUS get_panel_off_boot_request(uint8_t *is_panel_off_boot_request)
{
	if (pm_check_back_ground_active() == 0)
		// others
		*is_panel_off_boot_request = 0;
	else
		// Panel Off Boot
		*is_panel_off_boot_request = 1;

	return E_VBO_FRAME_OK;
}

static enum E_VBO_FRAME_STATUS get_bcm_fwupdate_status(uint8_t *is_bcm_fwupate_update_request)
{
	int32_t load_ret = -1;
	int32_t bcm_file_size = -1;
	char *bcm_file_val = NULL;

	bcm_file_size = ext4_f_size_mmc(UENV_PART_NAME, BCM_STATUS_FILE_PATH);
	if (bcm_file_size == -1) {
		UBOOT_INFO("File not found %s\n", BCM_STATUS_FILE_PATH);
		*is_bcm_fwupate_update_request = BCM_UPDATE_ENABLE;
		return E_VBO_FRAME_OK;
	}
	bcm_file_val = (char *)malloc(bcm_file_size);
	if (bcm_file_val == NULL) {
		UBOOT_ERROR("bcm_file_val malloc error\n");
		return E_VBO_FRAME_MALLOC_NG;
	}
	load_ret = ext4_f_load_mmc(UENV_PART_NAME, BCM_STATUS_FILE_PATH, bcm_file_val);
	if (load_ret == -1) {
		UBOOT_ERROR("FileOpen_Error %s\n", BCM_STATUS_FILE_PATH);
		free(bcm_file_val);
		return E_VBO_FRAME_BCM_FILEOPEN_NG;
	}

	if (strncmp(bcm_file_val, BCM_UPDATE_DISABLE_FLAG, BCM_UPDATE_FLAG_SIZE) == 0)
		*is_bcm_fwupate_update_request = BCM_UPDATE_DISABLE;
	else
		*is_bcm_fwupate_update_request = BCM_UPDATE_ENABLE;

	free(bcm_file_val);
	return E_VBO_FRAME_OK;
}
