/*
 * Copyright 2020,2021,2022 Sony Corporation
 */

#include <malloc.h>
#include <common.h>
#include <fs.h>
#include <utility.h>
#include <debug_impl.h>
#include <partition.h>
#include <sony_fs.h>

int ext4_f_generic_size(const char *device_name, const char *storage_info, const char *part_name, const char *file_name)
{
	loff_t size = 0;

	if (fs_set_blk_dev(device_name, storage_info, FS_TYPE_EXT)) {
		UBOOT_ERROR("partition select failure.\n");
		return -1;
	}

	if (fs_size(file_name, &size) < 0) {
		UBOOT_DEBUG("file size failure.\n");
		return -1;
	}

	return (size > 0) ? size : -1;
}

int ext4_f_generic_load(const char *device_name, const char *storage_info, const char *part_name, const char *file_name, char *buf)
{
	int ret = -1;
	loff_t readsize = 0;

	if (fs_set_blk_dev(device_name, storage_info, FS_TYPE_EXT)) {
		UBOOT_ERROR("partition select failure.\n");
		return -1;
	}

	if (fs_exists(file_name) == 0) {
		UBOOT_DEBUG("%s file not found.\n", file_name);
		return -1;
	}

	if (fs_set_blk_dev(device_name, storage_info, FS_TYPE_EXT)) {
		UBOOT_ERROR("partition select failure.\n");
		return -1;
	}

	ret = fs_read(file_name, (unsigned long)buf, 0, 0, &readsize);
	if (ret < 0) {
		UBOOT_ERROR("Read file failure.\n");
		return -1;
	}

	return readsize;
}

int ext4_f_size(const char *part_name, const char *file_name)
{
	int ret = -1;
	unsigned int device = 0;
	char device_name[STORAGE_DEVICE_BUF_SIZE] = { 0 };
	char storage_info[STORAGE_INDEX_BUF_SIZE] = { 0 };

	ret = sys_get_boot_device(device_name, sizeof(device_name));
	if (ret < 0) {
		UBOOT_ERROR("Get booting device name failure, Unknown storage device.\n");
		return -1;
	}

	ret = sys_get_storage_info(device, (char *)part_name, storage_info);
	if (ret < 0) {
		UBOOT_ERROR("sys_get_storage_info failure.\n");
		return -1;
	}

	return ext4_f_generic_size(device_name, storage_info, part_name, file_name);
}

int ext4_f_load(const char *part_name, const char *file_name, char *buf)
{
	int ret = -1;
	unsigned int device = 0;
	char device_name[STORAGE_DEVICE_BUF_SIZE] = { 0 };
	char storage_info[STORAGE_INDEX_BUF_SIZE] = { 0 };

	ret = sys_get_boot_device(device_name, sizeof(device_name));
	if (ret < 0) {
		UBOOT_ERROR("Get booting device name failure, Unknown storage device.\n");
		return -1;
	}

	ret = sys_get_storage_info(device, (char *)part_name, storage_info);
	if (ret < 0) {
		UBOOT_ERROR("sys_get_storage_info failure.\n");
		return -1;
	}

	return ext4_f_generic_load(device_name, storage_info, part_name, file_name, buf);
}

int ext4_f_size_mmc(const char *part_name, const char *file_name)
{
	int ret = -1;
	unsigned int device = 0;
	disk_partition_t info;
	struct blk_desc *blk_dev_desc = NULL;
	int part_num;
	unsigned int partition_num;
	char partition_name[PART_NAME_LEN] = { 0 };
	char storage_info[STORAGE_INDEX_BUF_SIZE] = { 0 };

	blk_dev_desc = blk_get_dev("mmc", device);
	if (get_boot_device() == EN_BOOT_DEVICE_USB) {
		part_num = part_info_search(blk_dev_desc, part_name, &info);
		/* slot a is used on usb boot */
		if (part_num <= 0) {
			snprintf(partition_name, PART_NAME_LEN - 1, "%s_a", part_name);
			part_num = part_info_search(blk_dev_desc, partition_name, &info);
		}
		if (part_num > 0) {
			partition_num = part_num;
			ret = 0;
		}
	} else {
		ret = get_partition_info(blk_dev_desc, part_name, NULL, &partition_num);
	}
	if (ret != 0) {
		UBOOT_DEBUG("get_partition_info failure.\n");
		return -1;
	}

	ret = snprintf(storage_info, (size_t)STORAGE_INDEX_BUF_SIZE, "%u:%x", device, partition_num);
	if (ret <= 0) {
		UBOOT_DEBUG("Error: storage_info fill data failure.\n");
		return -1;
	}

	return ext4_f_generic_size("mmc", storage_info, part_name, file_name);
}

int ext4_f_load_mmc(const char *part_name, const char *file_name, char *buf)
{
	int ret = -1;
	unsigned int device = 0;
	disk_partition_t info;
	struct blk_desc *blk_dev_desc = NULL;
	int part_num;
	unsigned int partition_num;
	char partition_name[PART_NAME_LEN] = { 0 };
	char storage_info[STORAGE_INDEX_BUF_SIZE] = { 0 };

	blk_dev_desc = blk_get_dev("mmc", device);
	if (get_boot_device() == EN_BOOT_DEVICE_USB) {
		part_num = part_info_search(blk_dev_desc, part_name, &info);
		/* slot a is used on usb boot */
		if (part_num <= 0) {
			snprintf(partition_name, PART_NAME_LEN - 1, "%s_a", part_name);
			part_num = part_info_search(blk_dev_desc, partition_name, &info);
		}
		if (part_num > 0) {
			partition_num = part_num;
			ret = 0;
		}
	} else {
		ret = get_partition_info(blk_dev_desc, part_name, NULL, &partition_num);
	}
	if (ret != 0) {
		UBOOT_DEBUG("get_partition_info failure.\n");
		return -1;
	}

	ret = snprintf(storage_info, (size_t)STORAGE_INDEX_BUF_SIZE, "%u:%x", device, partition_num);
	if (ret <= 0) {
		UBOOT_DEBUG("Error: storage_info fill data failure.\n");
		return -1;
	}

	return ext4_f_generic_load("mmc", storage_info, part_name, file_name, buf);
}
