/*
 * Copyright 2022 Sony Corporation
 */

#include <common.h>
#include <command.h>
#include <dm.h>
#include <dm/of_access.h>
#include <asm/gpio.h>
#include <mtk_dtbo.h>

#define STRING_BUFFER_SIZE 256

static bool search_str_with_delim(const char *str1, const char *str2, const char *delim)
{
	char buf[STRING_BUFFER_SIZE];
	char *p;

	strncpy(buf, str1, STRING_BUFFER_SIZE-1);
	buf[STRING_BUFFER_SIZE-1] = '\0';


	p = strtok(buf, delim);
	while (p != NULL) {
		if (strcmp(p, str2) == 0)
			return true;
		p = strtok(NULL, delim);
	}
	return false;
}

static int get_prop_gpio_base(void *fdt, unsigned int offset, uint32_t *base)
{
	const fdt32_t *cell;
	int len;

	cell = fdt_getprop(fdt, offset, "base", &len);
	if (!cell || len < sizeof(int)) {
		printf("Not found base property\n");
		return -EINVAL;
	}
	*base = fdt32_to_cpu(cell[0]);
	return 0;
}

int find_gpio_line_name(const char *gpio_line_name, uint32_t *gpio_num)
{
	unsigned long dtb_addr = 0;
	void *fdt;
	unsigned int offset = 0, offset_old = 0;
	int ret;

	// get dtb binary
	get_uboot_dtb_addr(&dtb_addr);
	fdt = (void *)dtb_addr;
	ret = fdt_check_header(fdt);
	if (ret != 0)
		return -EINVAL;

	// traverse fdt nodes
	for (offset = fdt_next_node(fdt, -1, NULL); offset >= 0;
			offset = fdt_next_node(fdt, offset, NULL)) {
		int count = 0;

		if (offset_old > offset)
			break;

		count = fdt_stringlist_count(fdt, offset, "gpio-line-names");
		if (count > 0) {
			for (int index = 0 ; index < count ; index++) {
				const char *str;

				str = fdt_stringlist_get(fdt, offset, "gpio-line-names", index, NULL);

				if (search_str_with_delim(str, gpio_line_name, "/")) {
					uint32_t base;

					ret = get_prop_gpio_base(fdt, offset, &base);
					if (ret < 0)
						return -EINVAL;

					*gpio_num = base + index;

					return 0;
				}
			}
		}

		offset_old = offset;
	}
	return -EINVAL;
}

#ifdef DEBUG_TEST_FIND_GPIO // test command
int test_find_gpio_line_name(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	char name[STRING_BUFFER_SIZE];
	uint32_t num = 0;
	int ret = 0;

	name[0] = '\0';
	if (argc < 2) {
		printf("name is empty\n");
		printf("name=%s\n", name);
	} else {
		strncpy(name, argv[1], STRING_BUFFER_SIZE);
		printf("name=%s\n", name);
	}

	ret = find_gpio_line_name(name, &num);
	printf("delim:/ GPIO name=%s  ret=%d num=%u;\n", name, ret, num);
	return 0;
}

U_BOOT_CMD(test_find_gpio_line_name, 2, 1, test_find_gpio_line_name,
	"Test GPIO Name get\n",
	"Get GPIO number by gpio-line-names\n"
);
#endif


