/*
 * Copyright 2022 Sony Corporation
 */

#include <common.h>
#include <audio_codec.h>
#include <div64.h>
#include <dm.h>
#include <i2c.h>
#include <i2s.h>
#include <sound.h>
#include <asm/gpio.h>

#include "prodconfig.h"
#include "damp_selector.h"

static int damp_selector_reset(struct udevice *dev, bool bEnable)
{
    struct damp_selector_priv *priv = dev_get_priv(dev);
    int ret = 0;

    if (dm_gpio_is_valid(&priv->amp_reset)) {
        ret = dm_gpio_set_value(&priv->amp_reset, bEnable);
    } else {
        ret = -EINVAL;
    }

    return ret;
}

static int damp_selector_set_params(struct udevice *dev, int interface, int rate,
                                    int mclk_freq, int bits_per_sample, uint channels)
{
    int ret;
    uint8_t read_buf[2] = {0x00};
    uint8_t buffer = 0x00;
    struct udevice *damp_dev;
    char damp_type[64] = {0x00};

    // Log output using dev_info is not available, so printf is used instead.
    printf("[DAMP selector] DAMP RESET HI\n");
    ret = damp_selector_reset(dev, 0);
    if (ret < 0) {
        dev_err(dev, "[DAMP selector] damp_selector_reset error:%d\n", ret);
    }
    mdelay(20);

    ret = get_product_config_value("Audio DAMP Type", damp_type);
    if (ret != 0) {
        dev_err(dev, "[DAMP selector] get_product_config_value error:%d\n", ret);
        return -EINVAL;
    } else if (strstr(damp_type, "AUTO_DET") == NULL) {
        printf("[DAMP selector] PCD is not covered by HW Scan:%s\n", damp_type);
        return 0;
    } else { // detect "AUTO_DET"
        printf("[DAMP selector] PCD is the target of HW Scan:%s\n", damp_type);
    }

    // RT9120S
    if (i2c_get_chip_for_busnum(1, 0x18, 1, &damp_dev) == 0) {
        ret = dm_i2c_read(damp_dev, 0x00, read_buf, 2);
        printf("[DAMP selector] RT9120S read i2c 0x00 ret:%d value:0x%04x expect:0x4303\n",
                ret, (unsigned int)((read_buf[0] << 8) | read_buf[1]));
    }

    // TAS5825
    if (i2c_get_chip_for_busnum(1, 0x4c, 1, &damp_dev) == 0) {
        dm_i2c_write(damp_dev, 0x00, &buffer, 1);
        dm_i2c_write(damp_dev, 0x7f, &buffer, 1);
        ret = dm_i2c_read(damp_dev, 0x67, &read_buf[0], 1);
        printf("[DAMP selector] TAS5825 read i2c 0x67 ret:%d value:0x%02x expect:0x95\n",
                ret, read_buf[0]);
    }

    // TAS5782
    if (i2c_get_chip_for_busnum(1, 0x48, 1, &damp_dev) == 0 ||
        i2c_get_chip_for_busnum(0, 0x48, 1, &damp_dev) == 0) {
        dm_i2c_write(damp_dev, 0x00, &buffer, 1);
        dm_i2c_write(damp_dev, 0x7f, &buffer, 1);
        ret = dm_i2c_read(damp_dev, 0x58, &read_buf[0], 1);
        printf("[DAMP selector] TAS5782 read i2c 0x58 ret:%d value:0x%02x expect:0x85\n",
                ret, read_buf[0]);
    }
    return 0;
}

static int damp_selector_probe(struct udevice *dev)
{
    struct damp_selector_priv *priv = dev_get_priv(dev);
    int ret = 0;

    ret = gpio_request_by_name(dev, "ampmute-gpios", 0, &priv->amp_mute, GPIOD_IS_OUT);
    if (!ret && dm_gpio_is_valid(&priv->amp_mute)) {
        dm_gpio_set_value(&priv->amp_mute, 1); // Set LOW
    }

    ret = gpio_request_by_name(dev, "ampreset-gpios", 0, &priv->amp_reset, GPIOD_IS_OUT);
    if (ret) {
        dev_err(dev, "[DAMP selector] get amp reset gpio fail %d\n", ret);
        return -EINVAL;
    }
    return 0;
}

static const struct audio_codec_ops damp_selector_ops = {
    .set_params = damp_selector_set_params,
};

static const struct udevice_id damp_selector_ids[] = {
    { .compatible = "sony,damp_selector" },
    { }
};

U_BOOT_DRIVER(damp_selector) = {
    .name       = "damp_selector",
    .id         = UCLASS_AUDIO_CODEC,
    .of_match   = damp_selector_ids,
    .probe      = damp_selector_probe,
    .ops        = &damp_selector_ops,
    .priv_auto_alloc_size = sizeof(struct damp_selector_priv),
};
