/*
 * Copyright 2022 Sony Corporation
 */

#include <common.h>
#include <command.h>
#include <dm.h>
#include <dm/of_access.h>
#include <asm/gpio.h>
#include <adc.h>
#include <mtk_dtbo.h>
#include <mtk-pm.h>
#include <boot_impl.h>
#include <powerkey_fdr_cmd.h>
#include <pwm.h>
#include <sony_gpio_util.h>

// definition for WLED
#define WLED_PWM_PERIOD 400000
#define WLED_PWM_DUTY 400000

static int lighting_wled_pwm(void)
{
	ofnode node;
	struct udevice *pwm;
	uint32_t wled_pwm_device_num;
	uint32_t wled_pwm_ch;
	int ret;

	node = ofnode_by_compatible(ofnode_null(), "sony,fdr");
	if (!ofnode_valid(node)) {
		printf("FDR device tree node not found\n");
		return -1;
	}

	ret = ofnode_read_u32(node, "wled-pwm-device-num", &wled_pwm_device_num);
	if (ret != 0) {
		printf("wled-pwm-device-num not found\n");
		return -1;
	}

	ret = ofnode_read_u32(node, "wled-pwm-ch", &wled_pwm_ch);
	if (ret != 0) {
		printf("wled-pwm-ch not found\n");
		return -1;
	}

	ret = uclass_get_device(UCLASS_PWM, wled_pwm_device_num, &pwm);
	if (ret != 0)
		printf("Not Found WLED pwm ret=%d\n", ret);

	pwm_set_config(pwm, wled_pwm_ch, WLED_PWM_PERIOD, WLED_PWM_DUTY);
	pwm_set_enable(pwm, wled_pwm_ch, true);

	return 0;
}

int do_check_powerkey_fdr(cmd_tbl_t *cmdtp, int flag, int argc, char *const argv[])
{
	struct udevice *dev;
	unsigned int data;
	int ret, cnt = 0, dev_cnt = 0;

	// FDR usecase is only from AC on
	if (pm_get_boot_reason() != PM_BR_AC) {
		printf("%s Skip powerkey check\n", __func__);
		return CMD_RET_SUCCESS;
	}
#if defined(ADC_PORT_POWERKEY_DEV) /* use a defined device in a header file */

	ret = uclass_get_device_by_name(UCLASS_ADC, ADC_PORT_POWERKEY_DEV, &dev);
	if (ret) {
		printf("ERR : %s, Unknown ADC device %s\n", __func__, ADC_PORT_POWERKEY_DEV);
		return CMD_RET_FAILURE;
	}
	dev_cnt = 1;

#else /* use auto detected device from device tree */

	/* find available adc devices */
	ret = uclass_first_device_err(UCLASS_ADC, &dev);
	if (ret) {
		printf("ERR : %s, No available ADC device\n", __func__);
		return CMD_RET_FAILURE;
	}
	/* check multiple adc device */
	do {
		dev_cnt++;
		ret = uclass_next_device(&dev);
		if (ret)
			return CMD_RET_FAILURE;
	} while (dev);

	if (dev_cnt > 1) {
		printf("ERR : %s, Has multiple ADC device\n", __func__);
		return CMD_RET_FAILURE;
	}
#endif

	ret = uclass_first_device_err(UCLASS_ADC, &dev);

	printf("Use device:%s port:%d for %s\n", dev->name, ADC_PORT_POWERKEY_NUM, __func__);

	ret = adc_channel_single_shot(dev->name, ADC_PORT_POWERKEY_NUM, &data);

	/* get adc value to check pressed and holded powerkey */

	if (data < ADC_PORT_POWERKEY_PRESSED_LEVEL_LOW) {
		printf("Powerkey is pressed\n");
	} else {
		printf("Powerkey is not pressed\n");
		return CMD_RET_SUCCESS;
	}

	/* get adc value to check holded powerkey */

	do {
		mdelay(1000);
		cnt++;
		ret = adc_channel_single_shot(dev->name, ADC_PORT_POWERKEY_NUM, &data);
	} while (cnt <= POWERKEY_FDR_CHECK_PERIOD_SEC && data < ADC_PORT_POWERKEY_PRESSED_LEVEL_LOW);

	if (cnt >= POWERKEY_FDR_CHECK_PERIOD_SEC) {

#if defined(POWERKEY_FDR_USE_RECOVERY)
		// LED
		int ret;

		printf("FDR LED\n");
		ret = lighting_wled_pwm();
		if (ret != 0)
			printf("LED lighting fail.\n");

		/* executing run_command("wipeuserdata", 0) */
		if (wipe_user_data() != 0) {
			printf("Write misc data to erase userdata failed!\n");
			return CMD_RET_FAILURE;
		}
		printf("Write misc data to erase userdata success\n");

		mdelay(8000); /* waiting for end of sync & indicate LED to user several seconds */
		printf("Finish to write and reset\n");
		do_reset(NULL, 0, 0, NULL);
#else
		/* executing run_command("wipeuserdataboot", 0) */
		if (wipe_user_data_in_uboot() != 0) {
			printf("Erase userdata in uboot failed!\n");
			return CMD_RET_FAILURE;
		}
		printf("Erase userdata in uboot success\n");
#endif

	}

	return CMD_RET_SUCCESS;
}

U_BOOT_CMD(check_pkey_fdr, 1, 1, do_check_powerkey_fdr,
	"Check Powerkey hold and Run FDR\n",
	"Check Powerkey hold and Run FDR when Powerkey is pressed and holded for a while\n"
);


