/*
 * Copyright 2020,2021,2022,2023 Sony Corporation
 */

#include <common.h>
#include <debug_impl.h>
#include <linux/ctype.h>
#include <malloc.h>
#include <sony_fs.h>
#include <prodconfig.h>

static struct sony_pcd_info *pcd_info;
static int pcd_info_index;

static int get_data(const char *key, const char *buf, char *value)
{
	char *find = NULL;
	char key_str[128] = { 0 };
	int i = 0;

	if ((key == NULL) || (buf == NULL) || (value == NULL)) {
		UBOOT_ERROR("parameter error.\n");
		return -1;
	}

	snprintf(key_str, sizeof(key_str) - 1, "\"%s\"", key);
	find = strstr(buf, key_str);
	if (find == NULL) {
		UBOOT_DEBUG("can't find %s key.\n", key);
		return -1;
	}

	// search ':'(0x3a)
	find = strchr(find, 0x3a);
	if (find == NULL) {
		UBOOT_DEBUG("syntax error.\n");
		return -1;
	}

	// search '"'(0x22)
	find = strchr(find, 0x22);
	if (find == NULL) {
		UBOOT_DEBUG("syntax error.\n");
		return -1;
	}

	// check '"'(0x22), 'LF'(0x0a)
	while ((*(++find) != 0x22) && (*find != 0x0a) && (*find != 0x00)) {
		value[i++] = *find;
		if (i == MAX_PRODUCT_CONFIG_LENGTH) {
			value[i-1] = '\0';
			break;
		}
	}
	return 0;
}

static int open_and_selection(const char *filename, const char *key, char *value)
{
	static bool show_once = true;
	char *buf = NULL;
	int size = -1;
	int ret = 0;
	bool is_persist = true;
	const char *part_name = NULL, *file_path = NULL;
	char persist_filename[64] = { 0 };

	if ((filename == NULL) || (key == NULL) || (value == NULL)) {
		UBOOT_ERROR("parameter error.\n");
		return -1;
	}

	snprintf(persist_filename, sizeof(persist_filename) - 1, "sony/%s", filename);
	file_path = persist_filename;
	size = ext4_f_size_mmc(PERSIST_PART_NAME, file_path);
	if (size == -1) {
		file_path = filename;
		size = ext4_f_size_mmc(SONY_CONFIG_PART_NAME, file_path);
		if  (size != -1) {
			if (show_once) {
				printf("read PCD from %s\n", SONY_CONFIG_PART_NAME);
				show_once = false;
			}
		} else {
			UBOOT_DEBUG("file size error.\n");
			return -1;
		}
		is_persist = false;
	} else {
		if (show_once) {
			printf("overlay PCD from %s\n", PERSIST_PART_NAME);
			show_once = false;
		}
	}

	buf = malloc(size+1);
	if (!buf) {
		UBOOT_ERROR("malloc error\n");
		return -1;
	}

	memset(buf, 0, size+1);
	part_name = is_persist ? PERSIST_PART_NAME : SONY_CONFIG_PART_NAME;

	ret = ext4_f_load_mmc(part_name, file_path, buf);
	if (ret > 0)
		ret = get_data(key, (const char *)buf, value);

	free(buf);
	return ret;
}

int pcd_info_start(void)
{
	if (pcd_info == NULL) {
		pcd_info = (struct sony_pcd_info *)malloc(sizeof(struct sony_pcd_info) * MAX_PRODUCT_CONFIG_ENTRY);
		assert(pcd_info != NULL);
		memset(pcd_info, 0, sizeof(struct sony_pcd_info) * MAX_PRODUCT_CONFIG_ENTRY);
	}
	return 0;
}

void pcd_info_end(void)
{
	if (pcd_info != NULL) {
		free(pcd_info);
		pcd_info = NULL;
		pcd_info_index = 0;
	}
}

int get_product_config_value(const char *key, char *value)
{
	int ret = 0, i = 0;
	char filename[64] = { 0 };
	char fullpath[128] = { 0 };

	// search pcd_info
	for (i = 0; i < pcd_info_index; i++) {
		if (strncmp(pcd_info[i].key, key, MAX_PRODUCT_CONFIG_LENGTH) == 0) {
			strncpy(value, pcd_info[i].value, MAX_PRODUCT_CONFIG_LENGTH);
			UBOOT_DEBUG("key=%s value=%s\n", key, value);
			return ret;
		}
	}

	snprintf(fullpath, sizeof(fullpath) - 1, "%s%s", PRODUCT_CONFIG_PATH, PRODUCT_CONFIG_MAP_FILE);
	// open map file and select json files
	ret = open_and_selection(fullpath, key, filename);
	if (ret == 0) {
		memset(fullpath, 0, sizeof(fullpath));
		snprintf(fullpath, sizeof(fullpath) - 1, "%s%s", PRODUCT_CONFIG_PATH, filename);
		// open json file and select key value
		ret = open_and_selection(fullpath, key, value);
		// add into pcd_info
		if ((ret == 0) && (pcd_info_index < MAX_PRODUCT_CONFIG_ENTRY)) {
			strncpy(pcd_info[pcd_info_index].key, key, MAX_PRODUCT_CONFIG_LENGTH-1);
			strncpy(pcd_info[pcd_info_index].value, value, MAX_PRODUCT_CONFIG_LENGTH-1);
			pcd_info_index++;
		}
	}
	UBOOT_DEBUG("key=%s value=%s\n", key, value);
	return ret;
}

