/*
 * Copyright 2020,2021,2022,2023 Sony Corporation
 */

#include <common.h>
#include <environment.h>
#include <utility.h>
#include <debug_impl.h>
#include <part.h>
#include <sony_fs.h>
#include <prodconfig.h>

#include <prodconfig_task.h>

#define BOOTDATA_PART_NAME "bootdata"
#define BOOTARGS_CONFIG_FILE "sony/autodet/Bootargsconfig.json"
#define KEY_ROSERIAL "RoSerial"

int set_sony_generation(char *config)
{
	int base_model_year = 2023;
	int config_model_year = 2023;

	config_model_year = strtoul(((config != NULL) ? config : "2023"), NULL, 0);
	snprintf(config, 3, "%d", config_model_year - base_model_year);

	return 0;
}

/*
 * parse bootargs config json, get prodconfig, and add to bootargs
 * # assuming json includes only string scalar at root level
 */
int parse_bootargs_config_json(const char *buf)
{
	char *start = NULL;
	char *end = NULL;
	char *now = NULL;
	bool parse_failed = false;

	start = strchr(buf, '{');
	end = strchr(buf, '}');

	if (start == NULL || end == NULL) {
		// incorrect json format.
		return -1;
	}

	now = start + 1;

	while (now < end) {
		char *key_start = NULL;
		char *key_end = NULL;
		char key[64] = { 0 };
		char *value_start = NULL;
		char *value_end = NULL;
		char value[64] = { 0 };
		char config[64] = { 0 };
		char bootargs_key[96] = { 0 };
		char bootargs_cfg[160] = { 0 };
		int ret = -1;

		key_start = strchr(now, '"');
		if (key_start == NULL) {
			// no more kv.
			break;
		}
		key_end = strchr(key_start+1, '"');
		if (key_end == NULL) {
			// incorrect kv define.
			parse_failed = true;
			break;
		}
		strncpy(key, key_start+1, key_end-key_start-1);

		value_start = strchr(key_end+1, '"');
		if (value_start == NULL) {
			// incorrect kv define.
			parse_failed = true;
			break;
		}
		value_end = strchr(value_start+1, '"');
		if (value_end == NULL) {
			// incorrect kv define.
			parse_failed = true;
			break;
		}
		strncpy(value, value_start+1, value_end-value_start-1);

		ret = get_product_config_value(key, config);

		if (strcmp(value, "vendor.extended_core.sony.generation") == 0)
			set_sony_generation(config);

		if (ret == 0) {
			snprintf(bootargs_key, sizeof(bootargs_key)-1, "androidboot.%s", value);
			snprintf(bootargs_cfg, sizeof(bootargs_cfg)-1, "%s=%s", bootargs_key, config);
			ret = add_bootargs(bootargs_key, bootargs_cfg, 0);
			if (ret != 0)
				UBOOT_ERROR("failed to add bootargs %s.\n", bootargs_cfg);
		} else {
			UBOOT_ERROR("failed to get prodconfig %s.\n", key);
		}

		now = value_end + 1;
	}

	if (parse_failed)
		return -1;

	return 0;
}

int prodconfig_task(void)
{
	int size = -1;
	char *buf = NULL;
	int ret = 0;

	size = ext4_f_size(BOOTDATA_PART_NAME, BOOTARGS_CONFIG_FILE);
	if (size == -1) {
		UBOOT_ERROR("failed to get file size of bootargs config file.\n");
		return -1;
	}

	buf = malloc(size+1);
	if (!buf) {
		UBOOT_ERROR("failed to malloc for bootargs config file.\n");
		return -1;
	}

	memset(buf, 0, size+1);
	ret = ext4_f_load(BOOTDATA_PART_NAME, BOOTARGS_CONFIG_FILE, buf);
	if (ret <= 0) {
		UBOOT_ERROR("failed to load bootargs config file.\n");
		free(buf);
		return -1;
	}

	if (parse_bootargs_config_json(buf) != 0) {
		UBOOT_ERROR("failed to parse bootargs config json.\n");
		free(buf);
		return -1;
	}

	free(buf);
	return 0;
}

int add_serialno(void)
{
	char roserial[64] = { 0 };
	char bootargs_key[96] = { 0 };
	char bootargs_cfg[160] = { 0 };
	int ret = -1;

	ret = get_product_config_value(KEY_ROSERIAL, roserial);

	if (ret == 0) {
		strncpy(bootargs_key, "androidboot.serialno", sizeof(bootargs_key)-1);
		snprintf(bootargs_cfg, sizeof(bootargs_cfg)-1, "%s=%s", bootargs_key, roserial);
		ret = add_bootargs(bootargs_key, bootargs_cfg, 0);
		if (ret != 0)
			UBOOT_ERROR("failed to add bootargs %s.\n", bootargs_cfg);
		ret = env_set("serialno", roserial);
		if (ret != 0)
			UBOOT_ERROR("failed to add serialno into uenv.\n");
	} else {
		UBOOT_ERROR("failed to get prodconfig.\n");
	}

	return 0;
}

