/*
 * Copyright 2023 Sony Corporation
 */

#include <common.h>
#include <utility.h>
#include <debug_impl.h>
#include "../../../u-boot-2019.04/lib/libavb/libavb.h"
#include <mtk_ab.h>

#include <rbiread.h>

// from avb_slot_verify.c
/* Maximum size of a vbmeta image - 64 KiB. */
#define VBMETA_MAX_SIZE (64 * 1024)

// from avb_cmdline.h.
#define AVB_PART_NAME_MAX_SIZE 32

int rbiread(void)
{
	int ret = -1;
	const char *partition_name = "vbmeta";
	char full_partition_name[AVB_PART_NAME_MAX_SIZE];
	char *ab_suffix;
	uint8_t *vbmeta_buf = NULL;
	AvbVBMetaImageHeader vbmeta_header;
	char command[64] = { 0 };
	char bootargs_key[100] = { 0 };
	char bootargs_value[32] = { 0 };
	char bootargs_cfg[150] = { 0 };
	char display_version[50] = { 0 };

	// check slot suffix.
	enum PARTITION_SELECT ab_selection_result = get_ab_selection_result();

	if (ab_selection_result == PARTITION_SELECT_A) {
		ab_suffix = "_a";
	} else if (ab_selection_result == PARTITION_SELECT_B) {
		ab_suffix = "_b";
	} else {
		UBOOT_ERROR("slot suffix is not a or b\n");
		return -1;
	}

	// Construct full partition name e.g. system_a.
	if (!avb_str_concat(full_partition_name,
		sizeof(full_partition_name),
		partition_name,
		avb_strlen(partition_name),
		ab_suffix,
		avb_strlen(ab_suffix))) {
		UBOOT_ERROR("Partition name and suffix does not fit.\n");
		return -1;
	}

	// Read from partition.
	vbmeta_buf = avb_malloc(VBMETA_MAX_SIZE);
	if (vbmeta_buf == NULL) {
		UBOOT_ERROR("fail to allocate vbmeta buf.\n");
		return -1;
	}
	snprintf(command, sizeof(command)-1, "partition read mmc 0 %s 0x%p 0x%X",
			full_partition_name, vbmeta_buf, VBMETA_MAX_SIZE);
	ret = run_command(command, 0);
	if (ret != 0) {
		UBOOT_ERROR("fail to read vbmeta from partition.\n");
		free(vbmeta_buf);
		return -1;
	}

	// Byteswap the header.
	avb_vbmeta_image_header_to_host_byte_order((AvbVBMetaImageHeader *)vbmeta_buf,
			&vbmeta_header);

	// Set to android boot prop for rollback_index
	snprintf(bootargs_key, sizeof(bootargs_key)-1,
			"androidboot.vendor.extended_core.sony.build.package.rollback_index");
	snprintf(bootargs_value, sizeof(bootargs_value)-1, "%llu", vbmeta_header.rollback_index);
	snprintf(bootargs_cfg, sizeof(bootargs_cfg)-1, "%s=%s", bootargs_key, bootargs_value);
	ret = add_bootargs(bootargs_key, bootargs_cfg, 0);
	if (ret != 0) {
		UBOOT_ERROR("failed to add bootargs %s.\n", bootargs_cfg);
		free(vbmeta_buf);
		return -1;
	}

	// Set to android boot prop for version.display
	snprintf(display_version, sizeof(display_version)-1, "%c%c%c.%c%c%c.%c%c%c.%c",
			bootargs_value[0], bootargs_value[1], bootargs_value[2],
			bootargs_value[3], bootargs_value[4], bootargs_value[5],
			bootargs_value[6], bootargs_value[7], bootargs_value[8],
			bootargs_value[9]);
	snprintf(bootargs_key, sizeof(bootargs_key)-1,
			"androidboot.vendor.extended_core.sony.build.package.version.display");
	snprintf(bootargs_cfg, sizeof(bootargs_cfg)-1, "%s=%s", bootargs_key, display_version);
	ret = add_bootargs(bootargs_key, bootargs_cfg, 0);
	if (ret != 0) {
		UBOOT_ERROR("failed to add bootargs %s.\n", bootargs_cfg);
		free(vbmeta_buf);
		return -1;
	}

	free(vbmeta_buf);
	return 0;
}
