/*
 * Copyright 2022 Sony Corporation
 */
#include "sony_gpio_init.h"
#include <asm/gpio.h>
#include <linux/delay.h>
#include <time.h>
#include "autodet.h"
#include <common.h>
#include <part.h>
#include <cli.h>
#include <partition.h>
#include <command.h>
#include "iniutility.h"
#include "system_impl.h"
#include <dm.h>

// #define DEBUG_MODE // Skip because debug mode is not used


enum _sony_gpio_init_task {
	INPUT_PROCESS,
	OUTPUT_LOW_PROCESS,
	OUTPUT_HIGH_PROCESS
};

enum _sony_gpio_init_task sony_gpio_init_task;



static uint32_t run_gpio_setting_func(uint32_t _gpio_num)
{
	int ret;
#ifdef DEBUG_MODE
	int old_ret;
#endif

	ret = gpio_request(_gpio_num, "sony_gpio_init");
	if (ret < 0) {
		printf("[%s, %d] GPIO(%d) request Failed.\n", LOG_TAG, __LINE__, _gpio_num);
		return 0;
	}

#ifdef DEBUG_MODE
	old_ret = gpio_get_value(_gpio_num);
	if (old_ret < 0)
		printf("[%s, %d] GPIO(%d) old status get Failed.\n", LOG_TAG, __LINE__, _gpio_num);
#endif

	switch (sony_gpio_init_task) {
	case INPUT_PROCESS:
#ifdef DEBUG_MODE
		printf("[%s, %d] GPIO(%d)'Input' task.\n", LOG_TAG, __LINE__, _gpio_num);
#endif
		ret = gpio_direction_input(_gpio_num);
		if (ret)
			printf("[%s, %d] GPIO(%d) status initialize 'Input' Failed.\n", LOG_TAG, __LINE__, _gpio_num);
		break;

	case OUTPUT_LOW_PROCESS:
#ifdef DEBUG_MODE
		printf("[%s, %d] GPIO(%d)'Output-Low' task.\n", LOG_TAG, __LINE__, _gpio_num);
#endif
		ret = gpio_direction_output(_gpio_num, 0);
		if (ret)
			printf("[%s, %d] GPIO(%d) status initialize 'Output-Low' Failed.\n", LOG_TAG, __LINE__, _gpio_num);
		break;

	case OUTPUT_HIGH_PROCESS:
#ifdef DEBUG_MODE
		printf("[%s, %d] GPIO(%d)'Output-High' task.\n", LOG_TAG, __LINE__, _gpio_num);
#endif
		ret = gpio_direction_output(_gpio_num, 1);
		if (ret)
			printf("[%s, %d] GPIO(%d) status initialize 'Output-High' Failed.\n", LOG_TAG, __LINE__, _gpio_num);
		break;

	default:
#ifdef DEBUG_MODE
		printf("[%s, %d] GPIO(%d) status is not initialize.\n", LOG_TAG, __LINE__, _gpio_num);
#endif
		break;
	}

#ifdef DEBUG_MODE
	ret = gpio_get_value(_gpio_num);
	if (ret < 0) {
		printf("[%s, %d] GPIO(%d) status get Failed.\n", LOG_TAG, __LINE__, _gpio_num);
	} else {
		printf("[%s, %d] GPIO(%d) status get Passed. value=[%d]\n", LOG_TAG, __LINE__, _gpio_num, ret);
		if (ret != old_ret)
			printf("[%s, %d] GPIO(%d)Value is [%d] => [%d]\n", LOG_TAG, __LINE__, _gpio_num, old_ret, ret);
	}
#endif
	ret = gpio_free(_gpio_num);
	if (ret < 0)
		printf("[%s, %d] GPIO(%d) free Failed.\n", LOG_TAG, __LINE__, _gpio_num);

	return 0;
}

static int parse_devicetree(struct udevice *dev)
{
	ofnode node;
	int i;
	int ret = -1;
	char gpio_init_buf[SONY_GPIO_SIZE];
	struct sony_gpio_init_state priv;

	node = ofnode_by_compatible(ofnode_null(), "sony,sony_gpio_init");
	if (!ofnode_valid(node)) {
		printf("[%s, %d] dts node is not found.\n", LOG_TAG, __LINE__);
		return -EINVAL;
	}

	for (i = 0; i < MAX_SONY_GPIO_SEARCH_NUM; i++) {
		snprintf(gpio_init_buf, sizeof(gpio_init_buf), "gpio-%d-input", i);
		ret = ofnode_read_u32(node, gpio_init_buf, &priv.gpio_num);
		if (ret == 0) {
			sony_gpio_init_task = INPUT_PROCESS;
			run_gpio_setting_func(priv.gpio_num);
			continue;
		}

		snprintf(gpio_init_buf, sizeof(gpio_init_buf), "gpio-%d-output-low", i);
		ret = ofnode_read_u32(node, gpio_init_buf, &priv.gpio_num);
		if (ret == 0) {
			sony_gpio_init_task = OUTPUT_LOW_PROCESS;
			run_gpio_setting_func(priv.gpio_num);
			continue;
		}

		snprintf(gpio_init_buf, sizeof(gpio_init_buf), "gpio-%d-output-high", i);
		ret = ofnode_read_u32(node, gpio_init_buf, &priv.gpio_num);
		if (ret == 0) {
			sony_gpio_init_task = OUTPUT_HIGH_PROCESS;
			run_gpio_setting_func(priv.gpio_num);
			continue;
		}

#ifdef DEBUG_MODE
		printf("[%s, %d]  GPIO(%d) dtsi prop is not found.\n", LOG_TAG, __LINE__, i);
#endif
	}

	return 0;
}

static int sony_gpio_init_probe(struct udevice *dev)
{
	printf("[%s, %d] Probed.\n", LOG_TAG, __LINE__);
	return 0;
}

static const struct udevice_id sony_gpio_init_ids[] = {
	{ .compatible = "sony,sony_gpio_init" },
	{ }
};

U_BOOT_DRIVER(sony_gpio_init) = {
	.name = "sony_gpio_init",
	.id = UCLASS_GPIO,
	.of_match = sony_gpio_init_ids,
	.ofdata_to_platdata = parse_devicetree,
	.probe = sony_gpio_init_probe,
	.priv_auto_alloc_size = sizeof(struct sony_gpio_init_state),
};

/*
 * sony_gpio_init() - Initialize gpio.
 * Describe and set the property in device tree as follows.
 *  For Input:
 *  - input-gpio-"gpio num" = <"gpio num">;
 *  For Output-Low:
 *  - output-low-gpio-"gpio num" = <"gpio num">;
 *  For Output-High:
 *  - output-high-gpio-"gpio num" = <"gpio num">;
 *
 * e.g.
 * In case of controlling GPIO1:
 *  sony_gpio_init: sony_gpio_init {
 *    compatible = "sony,sony_gpio_init";
 *    gpio-1-input = <1>;
 *  };
 */
int sony_gpio_init(void)
{
	struct udevice dummydev;

	printf("[%s] %s start.\n", LOG_TAG, __func__);
	parse_devicetree(&dummydev);
	printf("[%s] %s done.\n", LOG_TAG, __func__);
	return 0;
}

/*
 * sony_gpio_get_all() - get all gpio status.
 *
 * This function is for debugging only.
 * Not used during normal startup.
 */
int sony_gpio_get_all(void)
{
	int ret;
	uint32_t i;

	printf("[%s] %s start.\n", LOG_TAG, __func__);

	for (i = 0; i < MAX_SONY_GPIO_SEARCH_NUM; i++) {
		ret = gpio_free(i);
#ifdef DEBUG_MODE
		if (ret < 0)
			printf("[%s, %d] GPIO(%d) free Failed.\n", LOG_TAG, __LINE__, i);
#endif

		ret = gpio_request(i, "sony_gpio_init");
		if (ret < 0) {
#ifdef DEBUG_MODE
			printf("[%s, %d] GPIO(%d) request Failed.\n", LOG_TAG, __LINE__, i);
#endif
			continue;
		}

		ret = gpio_get_value(i);
		if (ret != -1)
			printf("[%s, %d] GPIO(%d) value=[%d]\n", LOG_TAG, __LINE__, i, ret);

		ret = gpio_free(i);
#ifdef DEBUG_MODE
		if (ret < 0)
			printf("[%s, %d] GPIO(%d) free Failed.\n", LOG_TAG, __LINE__, i);
#endif
	}

	printf("[%s] %s done.\n", LOG_TAG, __func__);
	return 0;
}
