/* SPDX-License-Identifier: (GPL-2.0-only OR BSD-3-Clause) */
/*
 * Copyright (c) 2023 MediaTek Inc.
 */

#ifndef _PARTITION_H_
#define _PARTITION_H_
/******************************************************************************
 *
 * Partition layer is generated by Mediatek-internal.
 * Users can use this layer of commands instead of native flash commands. (ex: mmc read)
 * The main difference of this layer and native flash commands are as follows,
 * 1. One command apllies to all flash type, user only needs to replace 'str_interface'
 * 2. Unit of size is different.
 *    Unit of partition layer is byte.
 *    Unit of native flash commands is block. (for eMMC, 1 block = 512 bytes in general)
 * 3. It also provides user reading & writing at specific start position in the partition
 *    Unit of specific start position in the partition is byte
 * 4. bootpartition API also pull type of device to the abstraction layer, user does not need to care which device is currently
 *
*****************************************************************************/


/******************************************************************************
 * One-PKG definition
*****************************************************************************/
/* define size of one-PKG header & sub-header */
#define HEADER_MAGIC_ID                  4
#define HEADER_RESERVED                  4
#define SUB_HEADER_NAME_LEN              8
#define SUB_HEADER_RESERVED              3

/* define chip id MT5896 & revision 3, using for special case MT5896 after E3 IC */
#define ASIC_MT5896_CHIP_ID              5896
#define ASIC_CHIP_REVISION_3             3

/* define _check_is_one_pkg return value */
#define PASS_IS_ONE_PKG                  0
#define PASS_NOT_ONE_PKG                 1
#define ERR_INVALID_ASIC_ID             -1
#define ERR_INVALID_ASIC_REVISION       -2
#define ERR_INVALID_ASIC_COMPLETE_NAME  -3
#define ERR_NO_CORR_PKG_IN_SUB_HEAD     -4
#define ERR_INVALID_PKG_CNT             -5


typedef struct
{
	char magic_id[HEADER_MAGIC_ID];
	unsigned int version;
	unsigned int count;
	unsigned char reserved[HEADER_RESERVED];
} one_pkg_main_header;

typedef struct
{
	char name[SUB_HEADER_NAME_LEN];
	unsigned int offset;
	unsigned int size;
	unsigned int id;
	unsigned int reserved[SUB_HEADER_RESERVED];
} one_pkg_sub_header;


/******************************************************************************
 *
 * get_boot_partition_info :  Search for a partition in boot partition by its partition name.
 *                            If the specified partition exists, save the partition information into variables.
 * target_name:               The partition name will be searched. First searching name is original 'target_name'.
 *                            If original 'target_name' is not found, search once with ab suffixed name.
 *                            ex: If partition name 'mboot' is not found, search partition name 'mboot_a' again.
 * target_info:               Output partition information including start LBA, size LBA.. will store in target_info
 * target_num:                Output partition index number on GPT will store in target_info
 *
 * return :                   0 = Success , 1 = Failure
 *
*****************************************************************************/
int get_boot_partition_info(const char *target_name, disk_partition_t *target_info, unsigned int *target_num);


/******************************************************************************
 *
 * get_partition_info :  Search for a partition by partition name.
 *                       If the specified partition exists, save the partition information into variables.
 * blk_dev_desc:         Device description
 * target_name:          The partition name will be searched. First searching name is original 'target_name'.
 *                       If original 'target_name' is not found, search once with ab suffixed name.
 *                       ex: If partition name 'mboot' is not found, search partition name 'mboot_a' again.
 * target_info:          Output partition information including start LBA, size LBA.. will store in target_info
 * target_num:           Output partition index number on GPT will store in target_info
 *
 * return :              0 = Success , 1 = Failure
 *
*****************************************************************************/
int get_partition_info(struct blk_desc *blk_dev_desc, const char *target_name, disk_partition_t *target_info, unsigned int *target_num);


/******************************************************************************
 *
 * do_partition_read:  Read the data on partition to DRAM.
 *                     Read 'str_size' bytes from 'str_partition_name' to address 'str_address'
 * str_interface:      Type of flash, Only support eMMC & USB & UFS now. ex: mmc, usb, ufs
 * str_device:         Number of device on flash, ex: 0, 1, ...
 * str_partition_name: Which partition you will read, ex: mboot_a
 * str_address:        The address to store the read data, ex: 0x22000000
 * str_size:           The size you will read. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 * str_start_byte_pos: The starting position of this partition you want to read. If 'pos' is omitted, 0 is used.
 *                     The unit of 'str_start_byte_pos' is byte in hexadecimal. ex: 0x800
 *
 * return :            0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_read(const char *str_interface, const char *str_device, const char *str_partition_name, const char *str_address, const char *str_size, const char *str_start_byte_pos);


/******************************************************************************
 *
 * do_partition_read_boot:  Read the data on boot partition to DRAM.
 *                          Read 'str_size' bytes bootdata from 'str_partition_number
 *                          'of 'str_device' on 'str_interface' to address 'str_address'
 * str_interface:           Type of flash, Only support eMMC & UFS now. ex: mmc, ufs
 * str_device:              Number of device on flash, ex: 0, 1, ...
 * str_partition_number:    Which boot partition you will read, ex: 0, 1...
 * str_address:             The address to store the read data, ex: 0x22000000
 * str_size:                The size you will read. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 * str_start_byte_pos:      The starting position of this partition you want to read. If 'pos' is omitted, 0 is used.
 *                          The unit of 'str_start_byte_pos' is byte in hexadecimal. ex: 0x800
 *
 * return :                 0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_read_boot(const char *str_interface, const char *str_device, const char *str_partition_number, const char *str_address, const char *str_size, const char *str_start_byte_pos);


/******************************************************************************
 *
 * do_partition_write: Write the data on DRAM to partition.
 *                     Write 'str_size' bytes to 'str_partition_name' from address 'str_address'
 * str_interface:      Type of flash, Only support eMMC & USB & UFS now. ex: mmc, usb, ufs
 * str_device:         Number of device on flash, ex: 0, 1, ...
 * str_partition_name: Which partition you will write, ex: mboot_a
 * str_address:        The address to store the data which will write into partition, ex: 0x22000000
 * str_size:           The size you will write. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 * str_start_byte_pos: The starting position of this partition you want to write. If 'pos' is omitted, 0 is used.
 *                     The unit of 'str_start_byte_pos' is byte in hexadecimal. ex: 0x800
 *
 * return :            0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_write(const char *str_interface, const char *str_device, const char *str_partition_name, const char *str_address, const char *str_size, const char *str_start_byte_pos);


/******************************************************************************
 *
 * do_partition_write_boot: Write the data on DRAM to boot partition.
 *                          Write 'str_size' bytes bootdata from address 'str_address'
 *                          to 'str_partition_number' of 'str_device' on 'str_interface'
 * str_interface:           Type of flash, Only support eMMC & UFS now. ex: mmc, ufs
 * str_device:              Number of device on flash, ex: 0, 1, ...
 * str_partition_number:    Which boot partition you will write, ex: 0, 1...
 * str_address:             The address to store the data which will write into boot partition, ex: 0x22000000
 * str_size:                The size you will write. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 * str_start_byte_pos:      The starting position of this partition you want to write. If 'pos' is omitted, 0 is used.
 *                          The unit of 'str_start_byte_pos' is byte in hexadecimal. ex: 0x800
 *
 * return :                 0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_write_boot(const char *str_interface, const char *str_device, const char *str_partition_number, const char *str_address, const char *str_size, const char *str_start_byte_pos);


/******************************************************************************
 *
 * do_partition_swrite: Write the sparse data on DRAM to partition.
 *                      If no sparse image in 'str_address', this API acts as 'do_partition_write' API.
 *                      Otherwise this API writes the whole partition and not references <str_size>
 * str_interface:       Type of flash, Only support eMMC & USB & UFS now. ex: mmc, usb, ufs
 * str_device:          Number of device on flash, ex: 0, 1, ...
 * str_partition_name:  Which partition you will write, ex: super
 * str_address:         The address to store the data which will write into partition, ex: 0x22000000
 * str_size:            The size you will write. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 *
 * return :             0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_swrite(const char *str_interface, const char *str_device, const char *str_partition_name, const char *str_address, const char *str_size);


/******************************************************************************
 *
 * do_partition_erase: Erase all data on partition.
 * str_interface:      Type of flash, Only support eMMC & UFS now. ex: mmc, ufs
 * str_device:         Number of device on flash, ex: 0, 1, ...
 * str_partition_name: Which partition you will erase, ex: mboot_a
 *
 * return :            0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_erase(const char *str_interface, const char *str_device, const char *str_partition_name);


/******************************************************************************
 *
 * do_partition_erase_boot: Erase all bootdata on boot partition.
 * str_interface:           Type of flash, Only support eMMC & UFS now. ex: mmc, ufs
 * str_device:              Number of device on flash, ex: 0, 1, ...
 * str_partition_number:    Which boot partition you will erase, ex: 0, 1
 *
 * return :                 0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_erase_boot(const char *str_interface, const char *str_device, const char *str_partition_number);


#if (CONFIG_AB_SIDELOAD == 1)
/******************************************************************************
 *
 * do_partition_erase: Clone partitions between A/B for ease of A/B boot test.
 * str_interface:      Type of flash, Only support eMMC & UFS now. ex: mmc, ufs
 * str_device:         Number of device on flash, ex: 0, 1, ...
 * str_action:         Clone action, ex: 0, 1, 2
 *
 * return :            0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_partition_clone(const char *str_interface, const char *str_device, const char *str_action);
#endif


/******************************************************************************
 *
 * do_boot_partition_write: Write the data on DRAM to partition.
 *                          Write 'str_size' bytes to 'str_partition_name' from address 'str_address'
 * str_partition_name:      Which partition you will write, ex: mboot_a
 * str_address:             The address to store the data which will write into partition, ex: 0x22000000
 * str_size:                The size you will write. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 * str_start_byte_pos:      The starting position of this partition you want to write. If 'pos' is omitted, 0 is used.
 *                          The unit of 'str_start_byte_pos' is byte in hexadecimal. ex: 0x800
 *
 * return :                 0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_boot_partition_write(const char *str_partition_name, const char *str_address, const char *str_size, const char *str_start_byte_pos);

/******************************************************************************
 *
 * do_boot_partition_read:  Read the data on partition to DRAM.
 *                          Read 'str_size' bytes from 'str_partition_name' to address 'str_address'
 * str_partition_name:      Which partition you will read, ex: mboot_a
 * str_address:             The address to store the read data, ex: 0x22000000
 * str_size:                The size you will read. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 * str_start_byte_pos:      The starting position of this partition you want to read. If 'pos' is omitted, 0 is used.
 *                          The unit of 'str_start_byte_pos' is byte in hexadecimal. ex: 0x800
 *
 * return :                 0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_boot_partition_read(const char *str_partition_name, const char *str_address, const char *str_size, const char *str_start_byte_pos);


/******************************************************************************
 *
 * do_boot_partition_swrite: Write the sparse data on DRAM to partition.
 *                           If no sparse image in 'str_address', this API acts as 'do_partition_write' API.
 *                           Otherwise this API writes the whole partition and not references <str_size>
 * str_partition_name:       Which partition you will write, ex: super
 * str_address:              The address to store the data which will write into partition, ex: 0x22000000
 * str_size:                 The size you will write. The unit of 'byte_count' is byte in hexadecimal. ex: 0x200
 *
 * return :                  0 = Success , 1 = Failure
 *
*****************************************************************************/
int do_boot_partition_swrite(const char *str_partition_name, const char *str_address, const char *str_size);


#endif

